import contextlib
from io import BytesIO
from tempfile import NamedTemporaryFile
from typing import Dict, List, Optional, Type, Union

import numpy as np
import torch

from lhotse.array import Array, TemporalArray
from lhotse.audio import AudioSource, Recording, RecordingSet
from lhotse.cut import CutSet, MonoCut, MultiCut
from lhotse.features import Features, FeatureSet
from lhotse.features.io import MemoryRawWriter
from lhotse.manipulation import Manifest
from lhotse.supervision import AlignmentItem, SupervisionSegment, SupervisionSet
from lhotse.utils import compute_num_frames, compute_num_samples, fastcopy


@contextlib.contextmanager
def as_lazy(manifest, suffix=".jsonl.gz"):
    """
    Context manager for converting eager manifests to lazy manifests.
    Intended for testing.
    """
    with NamedTemporaryFile(suffix=suffix) as f:
        manifest.to_file(f.name)
        f.flush()
        yield type(manifest).from_jsonl_lazy(f.name)


# noinspection PyPep8Naming
def DummyManifest(
    type_: Type, *, begin_id: int, end_id: int, with_data: bool = False
) -> Manifest:
    if type_ == RecordingSet:
        return RecordingSet.from_recordings(
            dummy_recording(idx, with_data=with_data) for idx in range(begin_id, end_id)
        )
    if type_ == SupervisionSet:
        return SupervisionSet.from_segments(
            dummy_supervision(idx) for idx in range(begin_id, end_id)
        )
    if type_ == FeatureSet:
        # noinspection PyTypeChecker
        return FeatureSet.from_features(
            dummy_features(idx, with_data=with_data) for idx in range(begin_id, end_id)
        )
    if type_ == CutSet:
        # noinspection PyTypeChecker
        return CutSet.from_cuts(
            dummy_cut(idx, supervisions=[dummy_supervision(idx)], with_data=with_data)
            for idx in range(begin_id, end_id)
        )


def dummy_recording(
    unique_id: int,
    duration: float = 1.0,
    sampling_rate: int = 16000,
    with_data: bool = False,
    source_format: str = "wav",
) -> Recording:
    num_samples = compute_num_samples(duration, sampling_rate)
    return Recording(
        id=f"dummy-recording-{unique_id:04d}",
        sources=[
            dummy_audio_source(
                sampling_rate=sampling_rate,
                num_samples=num_samples,
                with_data=with_data,
                format=source_format,
            )
        ],
        sampling_rate=sampling_rate,
        num_samples=num_samples,
        duration=duration,
    )


def dummy_audio_source(
    num_samples: int = 16000,
    sampling_rate: int = 16000,
    channels: Optional[List[int]] = None,
    with_data: bool = False,
    format: str = "wav",
) -> AudioSource:
    if channels is None:
        channels = [0]
    if not with_data:
        return AudioSource(
            type="command", channels=channels, source='echo "dummy waveform"'
        )
    else:
        import soundfile

        # generate 1kHz sine wave
        f_sine = 1000
        assert (
            f_sine < sampling_rate / 2
        ), f"Sine wave frequency {f_sine} exceeds Nyquist frequency {sampling_rate/2} for sampling rate {sampling_rate}"
        data = torch.sin(2 * np.pi * f_sine / sampling_rate * torch.arange(num_samples))

        # prepare multichannel data
        if len(channels) > 1:
            data = data.unsqueeze(0).expand(len(channels), -1).transpose(0, 1)
            # ensure each channel has different data for channel selection testing
            mults = torch.tensor([1 / idx for idx in range(1, len(channels) + 1)])
            data = data * mults

        # prepare source with the selected format
        binary_data = BytesIO()
        if format == "opus":
            # workaround for OPUS: soundfile supports OPUS as a subtype of OGG format
            soundfile.write(
                binary_data,
                data.cpu().numpy(),
                sampling_rate,
                format="OGG",
                subtype="OPUS",
                closefd=False,
            )
        else:
            soundfile.write(
                binary_data,
                data.cpu().numpy(),
                sampling_rate,
                format=format,
                closefd=False,
            )
        binary_data.seek(0)
        return AudioSource(
            type="memory", channels=channels, source=binary_data.getvalue()
        )


def dummy_multi_channel_recording(
    unique_id: int,
    duration: float = 1.0,
    sampling_rate: int = 16000,
    channel_ids: Optional[List[int]] = None,
    source_per_channel: bool = False,
    with_data: bool = False,
) -> Recording:
    num_samples = compute_num_samples(duration, sampling_rate)
    if channel_ids is None:
        channel_ids = [0, 1]
    if source_per_channel:
        sources = [
            dummy_audio_source(
                num_samples=num_samples,
                sampling_rate=sampling_rate,
                channels=channel_ids,
                with_data=with_data,
            )
        ]
    else:
        sources = [
            dummy_audio_source(
                num_samples=num_samples,
                sampling_rate=sampling_rate,
                channels=[i],
                with_data=with_data,
            )
            for i in channel_ids
        ]
    return Recording(
        id=f"dummy-multi-channel-recording-{unique_id:04d}",
        sources=sources,
        sampling_rate=sampling_rate,
        num_samples=num_samples,
        duration=duration,
    )


def dummy_alignment(
    text: str = "irrelevant", start: float = 0.0, duration: float = 1.0
) -> Dict[str, List[AlignmentItem]]:
    subwords = [
        text[i : i + 3] for i in range(0, len(text), 3)
    ]  # Create subwords of 3 chars
    dur = duration / len(subwords)
    alignment = [
        AlignmentItem(symbol=sub, start=start + i * dur, duration=dur)
        for i, sub in enumerate(subwords)
    ]
    return {"subword": alignment}


def dummy_supervision(
    unique_id: int,
    start: float = 0.0,
    duration: float = 1.0,
    channel: Union[int, List[int]] = 0,
    text: str = "irrelevant",
    alignment: Optional[Dict[str, List[AlignmentItem]]] = dummy_alignment(),
) -> SupervisionSegment:
    return SupervisionSegment(
        id=f"dummy-segment-{unique_id:04d}",
        recording_id=f"dummy-recording-{unique_id:04d}",
        start=start,
        duration=duration,
        channel=channel,
        text=text,
        speaker="irrelevant",
        language="irrelevant",
        gender="irrelevant",
        custom={"custom_field": "irrelevant"},
        alignment=alignment,
    )


def dummy_features(
    unique_id: int, start: float = 0.0, duration: float = 1.0, with_data: bool = False
) -> Features:
    # Note: with_data is ignored as this always has real data attached
    return Features(
        recording_id=f"dummy-recording-{unique_id:04d}",
        channels=0,
        start=start,
        duration=duration,
        type="fbank",
        num_frames=100,
        num_features=23,
        frame_shift=0.01,
        sampling_rate=16000,
        storage_type="lilcom_files",
        storage_path="test/fixtures/dummy_feats/storage",
        storage_key="dbf9a0ec-f79d-4eb8-ae83-143a6d5de64d.llc",
    )


def dummy_in_memory_features(
    unique_id: int,
    start: float = 0.0,
    duration: float = 1.0,
    sampling_rate: int = 16000,
    frame_shift: float = 0.01,
) -> Features:
    num_frames = compute_num_frames(duration, frame_shift, sampling_rate)
    num_features = 23
    data = np.random.rand(num_frames, num_features).astype(np.float32)
    bindata = MemoryRawWriter().write("dummy-features", data)
    return Features(
        recording_id=f"dummy-recording-{unique_id:04d}",
        channels=0,
        start=start,
        duration=duration,
        type="fbank",
        num_frames=num_frames,
        num_features=num_features,
        frame_shift=frame_shift,
        sampling_rate=sampling_rate,
        storage_type=MemoryRawWriter.name,
        storage_path="",
        storage_key=bindata,
    )


def dummy_multi_channel_features(
    unique_id: int,
    start: float = 0.0,
    duration: float = 1.0,
    channels: Optional[List[int]] = None,
) -> Features:
    if channels is None:
        channels = [0, 1]
    return Features(
        recording_id=f"dummy-multi-channel-recording-{unique_id:04d}",
        channels=channels,
        start=start,
        duration=duration,
        type="fbank",
        num_frames=100,
        num_features=23,
        frame_shift=0.01,
        sampling_rate=16000,
        storage_type="lilcom_files",
        storage_path="test/fixtures/dummy_feats/storage",
        storage_key="dbf9a0ec-f79d-4eb8-ae83-143a6d5de64d.llc",
    )


def dummy_temporal_array(
    start: float = 0.0,
    num_frames: int = 100,
    num_features: int = 23,
    frame_shift: float = 0.01,
) -> TemporalArray:
    data = np.random.rand(num_frames, num_features).astype(np.float32)
    return MemoryRawWriter().store_array(
        key="temporal-array-float32",
        value=data,
        frame_shift=frame_shift,
        temporal_dim=0,
        start=start,
    )


def dummy_array() -> Array:
    data = np.random.rand(128).astype(np.float32)
    return MemoryRawWriter().store_array("vector-float32", data)


def dummy_temporal_array_uint8(
    start: float = 0.0, num_frames: int = 100, frame_shift: float = 0.01
) -> TemporalArray:
    data = np.random.randint(0, 255, num_frames, dtype=np.uint8)
    return MemoryRawWriter().store_array(
        "temporal-array-int8",
        data,
        frame_shift=frame_shift,
        temporal_dim=0,
        start=start,
    )


def dummy_cut(
    unique_id: int,
    start: float = 0.0,
    duration: float = 1.0,
    recording_duration: float = 1.0,
    recording: Recording = None,
    features: Features = None,
    supervisions=None,
    with_data: bool = False,
):
    custom = {
        "custom_attribute": "dummy-value",
        "custom_attribute_other": "dummy-value-other",
    }
    if with_data:
        custom.update(
            {
                "custom_embedding": dummy_array(),
                "custom_features": dummy_temporal_array(start),
                "custom_recording": dummy_recording(
                    unique_id, duration=duration, with_data=True
                ),
                "custom_indexes": dummy_temporal_array_uint8(start=start),
            }
        )
    return MonoCut(
        id=f"dummy-mono-cut-{unique_id:04d}",
        start=start,
        duration=duration,
        channel=0,
        recording=recording
        if recording
        else dummy_recording(
            unique_id, duration=max(recording_duration, duration), with_data=with_data
        ),
        features=features
        if features
        else dummy_features(unique_id, with_data=with_data),
        supervisions=supervisions if supervisions is not None else [],
        custom=custom,
    )


def dummy_multi_cut(
    unique_id: int,
    start: float = 0.0,
    duration: float = 1.0,
    recording_duration: float = 1.0,
    recording: Recording = None,
    features: Features = None,
    supervisions: SupervisionSet = None,
    channel: Optional[List[int]] = None,
    source_per_channel: bool = False,
    with_data: bool = False,
):
    if channel is None:
        channel = [0, 1]
    if recording_duration < duration:
        recording_duration = duration
    return MultiCut(
        id=f"dummy-multi-cut-{unique_id:04d}",
        start=start,
        duration=duration,
        channel=channel,
        recording=recording
        if recording
        else dummy_multi_channel_recording(
            unique_id,
            duration=recording_duration,
            channel_ids=channel,
            with_data=with_data,
            source_per_channel=source_per_channel,
        ),
        features=features
        if features
        else dummy_multi_channel_features(unique_id, channels=channel),
        supervisions=supervisions if supervisions is not None else [],
    )


def remove_spaces_from_segment_text(segment):
    if segment.text is None:
        return segment
    return fastcopy(segment, text=segment.text.replace(" ", ""))



import io
import pickle
from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Dict, List, Type, Union

import numpy as np

from lhotse.features.io import FileIO
from lhotse.image.image import Image
from lhotse.utils import Pathlike, is_module_available


class ImageReader(metaclass=ABCMeta):
    """
    ``ImageReader`` defines the interface of how to load images from a particular storage backend.
    This backend could either be:

    - files on a local filesystem;
    - in-memory storage;
    - cloud storage;
    - etc.

    Each class inheriting from ``ImageReader`` must define:

    - the ``read()`` method, which defines the loading operation
      (accepts the ``key`` to locate the image in the storage and return it).
    - the ``name()`` property that is unique to this particular storage mechanism.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @abstractmethod
    def read(
        self,
        key: str,
    ) -> np.ndarray:
        ...


class ImageWriter(metaclass=ABCMeta):
    """
    ``ImageWriter`` defines the interface of how to store images in a particular storage backend.
    This backend could either be:

    - files on a local filesystem;
    - in-memory storage;
    - cloud storage;
    - etc.

    Each class inheriting from ``ImageWriter`` must define:

    - the ``write()`` method, which defines the storing operation
      (accepts a ``key`` used to place the ``value`` image in the storage);
    - the ``storage_path()`` property, which is either a directory for the files,
      or empty string for in-memory storage.
    - the ``name()`` property that is unique to this particular storage mechanism.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def storage_path(self) -> str:
        ...

    @abstractmethod
    def write(self, key: str, value: np.ndarray) -> str:
        ...

    def store_image(
        self,
        key: str,
        value: Union[str, np.ndarray, bytes],
    ) -> Image:
        """
        Store an image in the underlying storage and return a manifest
        describing how to retrieve it.

        :param key: An ID that uniquely identifies the image.
        :param value: The image to be stored. Can be:
            - A path to an image file
            - A numpy array with shape (height, width, channels)
            - Raw bytes of an image file
        :return: A manifest of type :class:`~lhotse.image.Image`
        """
        assert is_module_available(
            "PIL"
        ), "In order to store images, please run 'pip install pillow'"
        import PIL.Image

        # Handle different types of input values
        if isinstance(value, PIL.Image.Image):
            img = np.array(value)
        elif isinstance(value, str):
            # It's a path, load the image first
            img = np.array(PIL.Image.open(value))
        elif isinstance(value, np.ndarray):
            # It's a numpy array
            img = value
        elif isinstance(value, bytes):
            # It's raw bytes, load it into PIL first to get dimensions
            img = np.array(PIL.Image.open(io.BytesIO(value)))
        else:
            raise ValueError(f"Unsupported image value type: {type(value)}")

        height, width = img.shape[:2]
        storage_key = self.write(key, img)

        return Image(
            storage_type=self.name,
            storage_path=self.storage_path,
            storage_key=storage_key,
            width=width,
            height=height,
        )

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        ...


READER_BACKENDS = {}
WRITER_BACKENDS = {}


def available_storage_backends() -> List[str]:
    return sorted(set(READER_BACKENDS).intersection(WRITER_BACKENDS))


def register_reader(cls):
    """
    Decorator used to add a new ``ImageReader`` to Lhotse's registry.
    """
    READER_BACKENDS[cls.name] = cls
    return cls


def register_writer(cls):
    """
    Decorator used to add a new ``ImageWriter`` to Lhotse's registry.
    """
    WRITER_BACKENDS[cls.name] = cls
    return cls


def get_reader(name: str) -> Type[ImageReader]:
    """
    Find the available ``ImageReader`` implementation with the provided name.
    The name is the value of the ``name`` property in a concrete ``ImageReader`` implementation.
    """
    if name not in READER_BACKENDS:
        raise ValueError(
            f"No image reader backend with name '{name}' is available. "
            f"Available backends: {list(READER_BACKENDS.keys())}"
        )
    return READER_BACKENDS[name]


def get_writer(name: str) -> Type[ImageWriter]:
    """
    Find the available ``ImageWriter`` implementation with the provided name.
    The name is the value of the ``name`` property in a concrete ``ImageWriter`` implementation.
    """
    if name not in WRITER_BACKENDS:
        raise ValueError(
            f"No image writer backend with name '{name}' is available. "
            f"Available backends: {list(WRITER_BACKENDS.keys())}"
        )
    return WRITER_BACKENDS[name]


def get_memory_writer(name: str) -> Type[ImageWriter]:
    """
    Find the available in-memory ``ImageWriter`` implementation with the provided name.
    """
    writer = get_writer(name)
    assert is_in_memory(name), f"Storage type '{name}' is not an in-memory storage."
    return writer


def is_in_memory(storage_type: str) -> bool:
    """
    Determines whether a given storage_type represents an in-memory storage.
    """
    return storage_type == "pillow_memory"


@register_reader
class PillowReader(ImageReader):
    """
    Reads image files using Pillow from a directory on the local filesystem.
    ``storage_path`` corresponds to the directory path;
    ``storage_key`` for each image is the name of the file in that directory.
    """

    name = "pillow_files"

    def __init__(self, storage_path: Pathlike, *args, **kwargs):
        super().__init__()
        self.io = FileIO(storage_path)

    @property
    def storage_path(self) -> str:
        return self.io.storage_path

    def read(self, key: str, as_pil_image: bool = False):
        """Read the image file and return it as a numpy array."""
        assert is_module_available(
            "PIL"
        ), "In order to load images, please run 'pip install pillow'"
        import PIL.Image

        with self.io.open_fileobj(key, "r") as (f, input_path):
            img = PIL.Image.open(f)
            img.load()  # trigger read, pillow has lazy loading
        if as_pil_image:
            return img
        # Convert to numpy array (height, width, channels)
        return np.array(img)


@register_writer
class PillowWriter(ImageWriter):
    """
    Writes image files using Pillow to a directory on the local filesystem.
    ``storage_path`` corresponds to the directory path;
    ``storage_key`` for each image is the name of the file in that directory.
    """

    name = "pillow_files"

    def __init__(self, storage_path: Pathlike, *args, **kwargs):
        super().__init__()
        self.io = FileIO(storage_path)

    @property
    def storage_path(self) -> str:
        return self.io.storage_path

    def write(self, key: str, value: np.ndarray) -> str:
        """Write a numpy array as an image file."""
        assert is_module_available(
            "PIL"
        ), "In order to store images, please run 'pip install pillow'"
        import PIL.Image

        # Use PNG as the default format if no extension is provided
        if not Path(key).suffix:
            key = key + ".png"

        # Convert numpy array to PIL Image and save
        img = PIL.Image.fromarray(value)
        with self.io.open_fileobj(key, "w", add_subdir=True) as (f, output_path):
            img.save(f)

        # Include sub-directory in the key, e.g. "abc/abcdef.png"
        if not self.io.is_url:
            return "/".join(Path(output_path).parts[-2:])
        else:
            return key


@register_reader
class PillowInMemoryReader(ImageReader):
    """
    Reads images from memory storage using Pillow.
    """

    name = "pillow_memory"

    def __init__(self, *args, **kwargs):
        pass

    def read(self, raw_data: bytes, as_pil_image: bool = False) -> np.ndarray:
        """Read the image from the bytes in memory."""
        assert is_module_available(
            "PIL"
        ), "In order to load images, please run 'pip install pillow'"
        import PIL.Image

        # Decode from pickle, which contains either bytes or numpy array
        data = pickle.loads(raw_data)

        # If it's bytes, load with PIL
        if isinstance(data, bytes):
            img = PIL.Image.open(io.BytesIO(data))
            if as_pil_image:
                return img
            return np.array(img)
        # If it's already a numpy array, return it directly
        elif isinstance(data, np.ndarray):
            if as_pil_image:
                return PIL.Image.fromarray(data)
            return data
        elif isinstance(data, PIL.Image.Image):
            if as_pil_image:
                return data
            return np.array(data)
        else:
            raise ValueError(f"Unsupported data type: {type(data)}")


@register_writer
class PillowInMemoryWriter(ImageWriter):
    """
    Writes images to memory storage using Pillow.
    """

    name = "pillow_memory"

    def __init__(self, *args, **kwargs):
        pass

    @property
    def storage_path(self) -> None:
        return None

    def write(self, key: str, value: np.ndarray) -> bytes:
        """
        Store the image in memory. We're simply pickle-ing the numpy array.
        """
        return pickle.dumps(value)

    def close(self) -> None:
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import random

import numpy as np
import pytest
import torch


@pytest.fixture
def deterministic_rng(request):
    """
    Pytest fixture that ensures deterministic RNG behavior.
    After the test finishes, it restores the previous RNG state.

    Example usage::

        >>> def my_test(deterministic_rng):
        ...     x = torch.randn(10, 5)  # always has the same values

    You can also set random seed like this::

        >>> @pytest.mark.seed(1337)
        ... def my_test(deterministic_rng):
        ...     x = torch.randn(10, 5)

    .. note: Learn more about pytest fixtures setup/teardown here:
        https://docs.pytest.org/en/latest/how-to/fixtures.html#teardown-cleanup-aka-fixture-finalization
    """

    # The mechanism below is pytest's way of parameterizing fixtures.
    # We use that to optionally sed a different random seed than the default 0.
    # See: https://docs.pytest.org/en/7.1.x/how-to/fixtures.html#using-markers-to-pass-data-to-fixtures
    marker = request.node.get_closest_marker("seed")
    if marker is None:
        # Handle missing marker in some way...
        SEED = 0
    else:
        SEED = marker.args[0]

    torch_state = torch.get_rng_state()
    np_state = np.random.get_state()
    py_state = random.getstate()

    torch.manual_seed(SEED)
    np.random.seed(SEED)
    random.seed(SEED)

    yield SEED

    random.setstate(py_state)
    np.random.set_state(np_state)
    torch.set_rng_state(torch_state)



from dataclasses import dataclass
from io import BytesIO
from math import ceil, isclose
from pathlib import Path
from typing import Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import torch
from _decimal import ROUND_HALF_UP

from lhotse.audio.backend import get_current_audio_backend, info, save_audio
from lhotse.audio.source import AudioSource
from lhotse.audio.utils import (
    AudioLoadingError,
    DurationMismatchError,
    VideoInfo,
    get_audio_duration_mismatch_tolerance,
)
from lhotse.augmentation import (
    AudioTransform,
    Clipping,
    DereverbWPE,
    LoudnessNormalization,
    Narrowband,
    Resample,
    ReverbWithImpulseResponse,
    Speed,
    Tempo,
    Volume,
)
from lhotse.utils import (
    Pathlike,
    Seconds,
    SetContainingAnything,
    asdict_nonull,
    compute_num_samples,
    fastcopy,
    ifnone,
    perturb_num_samples,
    rich_exception_info,
)

Channels = Union[int, List[int]]


@dataclass
class Recording:
    """
    The :class:`~lhotse.audio.Recording` manifest describes the recordings in a given corpus.
    It contains information about the recording, such as its path(s), duration, the number of samples, etc.
    It allows to represent multiple channels coming from one or more files.

    This manifest does not specify any segmentation information or supervision such as the transcript or the speaker
    -- we use :class:`~lhotse.supervision.SupervisionSegment` for that.

    Note that :class:`~lhotse.audio.Recording` can represent both a single utterance (e.g., in LibriSpeech)
    and a 1-hour session with multiple channels and speakers (e.g., in AMI).
    In the latter case, it is partitioned into data suitable for model training using :class:`~lhotse.cut.Cut`.

    Internally, Lhotse supports multiple audio backends to read audio file.
    By default, we try to use libsoundfile, then torchaudio (with FFMPEG integration starting with torchaudio 2.1),
    and then audioread (which is an ffmpeg CLI wrapper).
    For sphere files we prefer to use sph2pipe binary as it can work with certain unique encodings such as "shorten".

    Audio backends in Lhotse are configurable. See:

    * :func:`~lhotse.audio.backend.available_audio_backends`
    * :func:`~lhotse.audio.backend.audio_backend`,
    * :func:`~lhotse.audio.backend.get_current_audio_backend`
    * :func:`~lhotse.audio.backend.set_current_audio_backend`
    * :func:`~lhotse.audio.backend.get_default_audio_backend`


    Examples

        A :class:`~lhotse.audio.Recording` can be simply created from a local audio file::

            >>> from lhotse import RecordingSet, Recording, AudioSource
            >>> recording = Recording.from_file('meeting.wav')
            >>> recording
            Recording(
                id='meeting',
                sources=[AudioSource(type='file', channels=[0], source='meeting.wav')],
                sampling_rate=16000,
                num_samples=57600000,
                duration=3600.0,
                transforms=None
            )

        This manifest can be easily converted to a Python dict and serialized to JSON/JSONL/YAML/etc::

            >>> recording.to_dict()
            {'id': 'meeting',
             'sources': [{'type': 'file',
               'channels': [0],
               'source': 'meeting.wav'}],
             'sampling_rate': 16000,
             'num_samples': 57600000,
             'duration': 3600.0}

        Recordings can be also created programatically, e.g. when they refer to URLs stored in S3 or somewhere else::

            >>> s3_audio_files = ['s3://my-bucket/123-5678.flac', ...]
            >>> recs = RecordingSet.from_recordings(
            ...     Recording(
            ...         id=url.split('/')[-1].replace('.flac', ''),
            ...         sources=[AudioSource(type='url', source=url, channels=[0])],
            ...         sampling_rate=16000,
            ...         num_samples=get_num_samples(url),
            ...         duration=get_duration(url)
            ...     )
            ...     for url in s3_audio_files
            ... )

        It allows reading a subset of the audio samples as a numpy array::

            >>> samples = recording.load_audio()
            >>> assert samples.shape == (1, 16000)
            >>> samples2 = recording.load_audio(offset=0.5)
            >>> assert samples2.shape == (1, 8000)

        See also: :class:`~lhotse.audio.recording.Recording`, :class:`~lhotse.cut.Cut`, :class:`~lhotse.cut.CutSet`.
    """

    id: str
    sources: List[AudioSource]
    sampling_rate: int
    num_samples: int
    duration: Seconds
    channel_ids: Optional[List[int]] = None
    transforms: Optional[List[Union[AudioTransform, Dict]]] = None

    def __post_init__(self):
        if self.channel_ids is None:
            self.channel_ids = sorted(
                cid for source in self.sources for cid in source.channels
            )
        assert (
            sum(source.has_video for source in self.sources) < 2
        ), "Lhotse does not currently support recordings with more than a single video stream."

    @property
    def video(self) -> Optional[VideoInfo]:
        s = self._video_source
        if s is None:
            return None
        return s.video

    @property
    def has_video(self) -> bool:
        return self._video_source is not None

    @property
    def _video_source(self) -> Optional[AudioSource]:
        for s in self.sources:
            if s.has_video:
                return s
        return None

    @property
    def is_in_memory(self) -> bool:
        return any(s.type == "memory" for s in self.sources)

    @property
    def is_placeholder(self) -> bool:
        return any(s.type == "shar" for s in self.sources)

    @property
    def num_channels(self) -> int:
        return len(self.channel_ids)

    @property
    def source_format(self) -> str:
        """Infer format of the audio sources.
        If all sources have the same format, return it.
        If sources have different formats, raise an error.
        """
        source_formats = list(set([s.format for s in self.sources]))

        if len(source_formats) == 1:
            # if all sources have the same format, return it
            return source_formats[0]
        else:
            # at the moment, we don't resolve different formats
            raise NotImplementedError(
                "Sources have different formats. Resolving to a single format not implemented."
            )

    @staticmethod
    def from_file(
        path: Pathlike,
        recording_id: Optional[Union[str, Callable[[Path], str]]] = None,
        relative_path_depth: Optional[int] = None,
        force_opus_sampling_rate: Optional[int] = None,
        force_read_audio: bool = False,
    ) -> "Recording":
        """
        Read an audio file's header and create the corresponding ``Recording``.
        Suitable to use when each physical file represents a separate recording session.

        .. caution::
            If a recording session consists of multiple files (e.g. one per channel),
            it is advisable to create the ``Recording`` object manually, with each
            file represented as a separate ``AudioSource`` object.

        :param path: Path to an audio file supported by libsoundfile (pysoundfile).
        :param recording_id: recording id, when not specified ream the filename's stem ("x.wav" -> "x").
            It can be specified as a string or a function that takes the recording path and returns a string.
        :param relative_path_depth: optional int specifying how many last parts of the file path
            should be retained in the ``AudioSource``. By default writes the path as is.
        :param force_opus_sampling_rate: when specified, this value will be used as the sampling rate
            instead of the one we read from the manifest. This is useful for OPUS files that always
            have 48kHz rate and need to be resampled to the real one -- we will perform that operation
            "under-the-hood". For non-OPUS files this input is undefined.
        :param force_read_audio: Set it to ``True`` for audio files that do not have any metadata
            in their headers (e.g., "The People's Speech" FLAC files).
        :return: a new ``Recording`` instance pointing to the audio file.
        """
        path = Path(path)
        recording_id = (
            path.stem
            if recording_id is None
            else recording_id(path)
            if callable(recording_id)
            else recording_id
        )
        audio_info = info(
            path,
            force_opus_sampling_rate=force_opus_sampling_rate,
            force_read_audio=force_read_audio,
        )
        if audio_info.video is not None:
            duration = audio_info.video.duration
            num_samples = compute_num_samples(duration, audio_info.samplerate)
        else:
            duration = audio_info.duration
            num_samples = audio_info.frames
        return Recording(
            id=recording_id,
            sampling_rate=audio_info.samplerate,
            num_samples=num_samples,
            duration=duration,
            sources=[
                AudioSource(
                    type="file",
                    channels=list(range(audio_info.channels)),
                    source=(
                        "/".join(path.parts[-relative_path_depth:])
                        if relative_path_depth is not None and relative_path_depth > 0
                        else str(path)
                    ),
                    video=audio_info.video,
                )
            ],
        )

    @staticmethod
    def from_bytes(
        data: bytes,
        recording_id: str,
    ) -> "Recording":
        """
        Like :meth:`.Recording.from_file`, but creates a manifest for a byte string with
        raw encoded audio data. This data is first decoded to obtain info such as the
        sampling rate, number of channels, etc. Then, the binary data is attached to the
        manifest. Calling :meth:`.Recording.load_audio` does not perform any I/O and
        instead decodes the byte string contents in memory.

        .. note:: Intended use of this method is for packing Recordings into archives
            where metadata and data should be available together
            (e.g., in WebDataset style tarballs).

        .. caution:: Manifest created with this method cannot be stored as JSON
            because JSON doesn't allow serializing binary data.

        :param data: bytes, byte string containing encoded audio contents.
        :param recording_id: recording id, unique string identifier.
        :return: a new ``Recording`` instance that owns the byte string data.
        """
        stream = BytesIO(data)
        audio_info = get_current_audio_backend().info(stream)
        return Recording(
            id=recording_id,
            sampling_rate=audio_info.samplerate,
            num_samples=audio_info.frames,
            duration=audio_info.duration,
            sources=[
                AudioSource(
                    type="memory",
                    channels=list(range(audio_info.channels)),
                    source=data,
                )
            ],
        )

    def move_to_memory(
        self,
        channels: Optional[Channels] = None,
        offset: Seconds = None,
        duration: Optional[Seconds] = None,
        format: Optional[str] = None,
    ) -> "Recording":
        """
        Read audio data and return a copy of the manifest with binary data attached.
        Calling :meth:`.Recording.load_audio` on that copy will not trigger I/O.

        If all arguments are left as defaults, we won't decode the audio and attach
        the bytes we read from disk/other source as-is.
        If ``channels``, ``duration``, or ``offset`` are specified, we'll decode the
        audio and re-encode it into ``format`` before attaching.
        The default format is FLAC, other formats compatible with torchaudio.save are
        also accepted.
        """

        if all(src.type == "memory" for src in self.sources):
            return self  # nothing to do

        def _aslist(x):
            if isinstance(x, int):
                return [x]
            return x

        # Case #1: no opts specified, read audio without decoding and move it in memory.
        if all(opt is None for opt in (channels, offset, duration)) or (
            (channels is None or _aslist(channels) == self.channel_ids)
            and (offset is None or isclose(offset, 0.0))
            and (duration is None or isclose(duration, self.duration))
        ):
            memory_sources = [
                AudioSource(
                    type="memory",
                    channels=old_source.channels,
                    source=open(old_source.source, "rb").read(),
                )
                for old_source in self.sources
            ]
            return fastcopy(self, sources=memory_sources)

        # Case #2: user specified some subset of the recording, decode audio,
        #          subset it, and encode it again but save in memory.
        audio = self.load_audio(
            channels=channels, offset=ifnone(offset, 0), duration=duration
        )
        stream = BytesIO()
        save_audio(stream, torch.from_numpy(audio), self.sampling_rate, format=format)
        channels = ifnone(channels, self.channel_ids)
        if isinstance(channels, int):
            channels = [channels]
        return Recording(
            id=self.id,
            sources=[
                AudioSource(
                    type="memory",
                    channels=channels,
                    source=stream.getvalue(),
                )
            ],
            sampling_rate=self.sampling_rate,
            num_samples=audio.shape[1],
            duration=ifnone(duration, self.duration),
        )

    def to_dict(self) -> dict:
        d = asdict_nonull(self)
        if self.transforms is not None:
            d["transforms"] = [
                t if isinstance(t, dict) else t.to_dict() for t in self.transforms
            ]
        return d

    def to_cut(self):
        """
        Create a Cut out of this recording --- MonoCut or MultiCut, depending on the
        number of channels.
        """
        from lhotse.cut import MonoCut, MultiCut

        cls = MonoCut if self.num_channels == 1 else MultiCut
        return cls(
            id=self.id,
            start=0.0,
            duration=self.duration,
            channel=self.channel_ids[0] if self.num_channels == 1 else self.channel_ids,
            recording=self,
        )

    @rich_exception_info
    def load_audio(
        self,
        channels: Optional[Channels] = None,
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
    ) -> np.ndarray:
        """
        Read the audio samples from the underlying audio source (path, URL, unix pipe/command).

        :param channels: int or iterable of ints, a subset of channel IDs to read (reads all by default).
        :param offset: seconds, where to start reading the audio (at offset 0 by default).
            Note that it is only efficient for local filesystem files, i.e. URLs and commands will read
            all the samples first and discard the unneeded ones afterwards.
        :param duration: seconds, indicates the total audio time to read (starting from ``offset``).
        :return: a numpy array of audio samples with shape ``(num_channels, num_samples)``.
        """

        assert offset <= self.duration, (
            f"Cannot load audio because the Recording's duration {self.duration}s "
            f"is smaller than the requested offset {offset}s."
        )

        # Micro-optimization for a number of audio loading cases:
        # if duration is very close to full recording,
        # just read everything, and we'll discard some samples at the end.
        orig_duration = duration
        if duration is not None and isclose(duration, self.duration, abs_tol=1e-3):
            duration = None

        if channels is None:
            channels = SetContainingAnything()
        else:
            channels = frozenset([channels] if isinstance(channels, int) else channels)
            recording_channels = frozenset(self.channel_ids)
            assert channels.issubset(recording_channels), (
                "Requested to load audio from a channel "
                "that does not exist in the recording: "
                f"(recording channels: {recording_channels} -- "
                f"requested channels: {channels})"
            )

        transforms = [
            tnfm if isinstance(tnfm, AudioTransform) else AudioTransform.from_dict(tnfm)
            for tnfm in self.transforms or []
        ]

        # Do a "backward pass" over data augmentation transforms to get the
        # offset and duration for loading a piece of the original audio.
        offset_aug, duration_aug = offset, duration
        for tfn in reversed(transforms):
            offset_aug, duration_aug = tfn.reverse_timestamps(
                offset=offset_aug,
                duration=duration_aug,
                sampling_rate=self.sampling_rate,
            )

        samples_per_source = []
        for source in self.sources:
            # Case: source not requested
            if not channels.intersection(source.channels):
                continue
            samples = source.load_audio(
                offset=offset_aug,
                duration=duration_aug,
                force_opus_sampling_rate=self.sampling_rate,
            )

            # Case: two-channel audio file but only one channel requested
            #       it might not be optimal to load all channels, but IDK if there's anything we can do about it
            channels_to_remove = [
                idx for idx, cid in enumerate(source.channels) if cid not in channels
            ]
            if channels_to_remove:
                samples = np.delete(samples, channels_to_remove, axis=0)
            samples_per_source.append(samples)

        # Stack all the samples from all the sources into a single array.
        audio = self._stack_audio_channels(samples_per_source)

        # We'll apply the transforms now (if any).
        for tfn in transforms:
            audio = tfn(audio, self.sampling_rate)

        if self.has_video:
            # It's possible the audio and video durations are quite mismatched.
            # We'll pad audio with zeroes or truncate audio to accomodate the video,
            # when it's available
            audio = assert_and_maybe_fix_num_samples(
                audio,
                offset=offset,
                duration=orig_duration,
                recording=self,
                tolerance=1e6,
                pad_mode="constant",
            )
        else:
            # Transformation chains can introduce small mismatches in the number of samples:
            # we'll fix them here, or raise an error if they exceeded a tolerance threshold.
            audio = assert_and_maybe_fix_num_samples(
                audio, offset=offset, duration=orig_duration, recording=self
            )

        return audio

    @rich_exception_info
    def load_video(
        self,
        channels: Optional[Channels] = None,
        offset: Seconds = 0.0,
        duration: Optional[Seconds] = None,
        with_audio: bool = True,
        force_consistent_duration: bool = True,
    ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
        """
        Read the video frames and audio samples from the underlying source (path, URL, unix pipe/command).

        :param channels: int or iterable of ints, a subset of channel IDs to read (reads all by default).
        :param offset: seconds, where to start reading the video (at offset 0 by default).
            Note that it is only efficient for local filesystem files, i.e. URLs and commands will read
            all the samples first and discard the unneeded ones afterwards.
        :param duration: seconds, indicates the total video time to read (starting from ``offset``).
        :param with_audio: bool, whether to load and return audio alongside video. True by default.
        :param force_consistent_duration: bool, if audio duration is different than video duration
            (as counted by ``num_frames / fps``), we'll either truncate or pad the audio with zeros.
            True by default.
        :return: a tuple of video tensor and optional audio tensor (or None).
        """

        assert self.has_video, f"Recording {self.id} has no video to load."

        assert offset <= self.duration, (
            f"Cannot load audio because the Recording's duration {self.duration}s "
            f"is smaller than the requested offset {offset}s."
        )

        for t in ifnone(self.transforms, ()):
            if isinstance(t, dict):
                assert t["name"] not in (
                    "Speed",
                    "Tempo",
                ), "Recording.load_video() does not support speed/tempo perturbation."
            else:
                assert not isinstance(
                    t, (Speed, Tempo)
                ), "Recording.load_video() does not support speed/tempo perturbation."

        if not with_audio:
            video, _ = self._video_source.load_video(
                offset=offset, duration=duration, with_audio=False
            )
            return video, None

        # Micro-optimization for a number of audio loading cases:
        # if duration is very close to full recording,
        # just read everything, and we'll discard some samples at the end.
        orig_duration = duration
        if duration is not None and isclose(duration, self.duration, abs_tol=1e-3):
            duration = None

        if channels is None:
            channels = SetContainingAnything()
        else:
            channels = frozenset([channels] if isinstance(channels, int) else channels)
            recording_channels = frozenset(self.channel_ids)
            assert channels.issubset(recording_channels), (
                "Requested to load audio from a channel "
                "that does not exist in the recording: "
                f"(recording channels: {recording_channels} -- "
                f"requested channels: {channels})"
            )

        transforms = [
            tnfm if isinstance(tnfm, AudioTransform) else AudioTransform.from_dict(tnfm)
            for tnfm in self.transforms or []
        ]

        # Do a "backward pass" over data augmentation transforms to get the
        # offset and duration for loading a piece of the original audio.
        offset_aug, duration_aug = offset, duration
        for tfn in reversed(transforms):
            offset_aug, duration_aug = tfn.reverse_timestamps(
                offset=offset_aug,
                duration=duration_aug,
                sampling_rate=self.sampling_rate,
            )

        samples_per_source = []
        video = None
        for source in self.sources:
            if source.has_video:
                video, samples = source.load_video(
                    offset=offset_aug,
                    duration=duration_aug,
                )
            else:
                samples = source.load_audio(offset=offset_aug, duration=duration_aug)

            # Case: source not requested (for audio, but it might be the only one with video)
            if not channels.intersection(source.channels):
                continue

            # Case: two-channel audio file but only one channel requested
            #       it might not be optimal to load all channels, but IDK if there's anything we can do about it
            channels_to_remove = [
                idx for idx, cid in enumerate(source.channels) if cid not in channels
            ]
            if channels_to_remove:
                samples = np.delete(samples, channels_to_remove, axis=0)
            samples_per_source.append(samples)

        assert video is not None

        # Stack all the samples from all the sources into a single array.
        audio = self._stack_audio_channels(samples_per_source)

        # We'll apply the transforms now (if any).
        for tfn in transforms:
            audio = tfn(audio, self.sampling_rate)

        if force_consistent_duration:
            # We want to keep audio and video duration identical by truncating/padding audio.
            audio = assert_and_maybe_fix_num_samples(
                audio,
                offset=offset,
                duration=video.shape[0] / self.video.fps,
                recording=self,
                # hack: "infinite" tolerance disables exceptions, i.e. 1min video and 1h audio => 1min audio
                tolerance=1e6,
                pad_mode="zero",
            )
        else:
            # Transformation chains can introduce small mismatches in the number of samples:
            # we'll fix them here, or raise an error if they exceeded a tolerance threshold.
            audio = assert_and_maybe_fix_num_samples(
                audio,
                offset=offset,
                duration=orig_duration,
                recording=self,
                pad_mode="reflect",
            )

        return video, torch.from_numpy(audio)

    def play_video(self):
        if self.has_video:
            from IPython.display import Video

            return Video(filename=self._video_source.source)

    def _stack_audio_channels(self, samples_per_source: List[np.ndarray]) -> np.ndarray:
        # There may be a mismatch in the number of samples between different channels. We
        # check if the mismatch is within a reasonable tolerance and if so, we pad
        # all channels to the length of the longest one.
        allowed_diff = int(
            compute_num_samples(
                get_audio_duration_mismatch_tolerance(),
                sampling_rate=self.sampling_rate,
            )
        )
        if len(samples_per_source) > 1:
            # Make all arrays 2D
            samples_per_source = [
                s[None, :] if s.ndim == 1 else s for s in samples_per_source
            ]
            max_samples = max(s.shape[1] for s in samples_per_source)
            for i, s in enumerate(samples_per_source):
                if max_samples - s.shape[1] <= allowed_diff:
                    s = np.pad(s, ((0, 0), (0, max_samples - s.shape[1])), "constant")
                    samples_per_source[i] = s
                else:
                    raise DurationMismatchError(
                        f"The mismatch between the number of samples in the "
                        f"different channels of the recording {self.id} is "
                        f"greater than the allowed tolerance {get_audio_duration_mismatch_tolerance()}."
                    )
            audio = np.concatenate(samples_per_source, axis=0)
        else:
            # shape: (n_channels, n_samples)
            audio = np.vstack(samples_per_source)
        return audio

    def _expected_num_samples(
        self, offset: Seconds, duration: Optional[Seconds]
    ) -> int:
        if offset == 0 and duration is None:
            return self.num_samples
        duration = duration if duration is not None else self.duration - offset
        return compute_num_samples(duration, sampling_rate=self.sampling_rate)

    def with_path_prefix(self, path: Pathlike) -> "Recording":
        return fastcopy(self, sources=[s.with_path_prefix(path) for s in self.sources])

    def with_video_resolution(self, width: int, height: int) -> "Recording":
        return fastcopy(
            self,
            sources=[
                s.with_video_resolution(width=width, height=height)
                for s in self.sources
            ],
        )

    def perturb_speed(self, factor: float, affix_id: bool = True) -> "Recording":
        """
        Return a new ``Recording`` that will lazily perturb the speed while loading audio.
        The ``num_samples`` and ``duration`` fields are updated to reflect the
        shrinking/extending effect of speed.

        :param factor: The speed will be adjusted this many times (e.g. factor=1.1 means 1.1x faster).
        :param affix_id: When true, we will modify the ``Recording.id`` field
            by affixing it with "_sp{factor}".
        :return: a modified copy of the current ``Recording``.
        """
        transforms = self.transforms.copy() if self.transforms is not None else []
        transforms.append(Speed(factor=factor))
        new_num_samples = perturb_num_samples(self.num_samples, factor)
        new_duration = new_num_samples / self.sampling_rate
        return fastcopy(
            self,
            id=f"{self.id}_sp{factor}" if affix_id else self.id,
            num_samples=new_num_samples,
            duration=new_duration,
            transforms=transforms,
        )

    def perturb_tempo(self, factor: float, affix_id: bool = True) -> "Recording":
        """
        Return a new ``Recording`` that will lazily perturb the tempo while loading audio.

        Compared to speed perturbation, tempo preserves pitch.
        The ``num_samples`` and ``duration`` fields are updated to reflect the
        shrinking/extending effect of tempo.

        :param factor: The tempo will be adjusted this many times (e.g. factor=1.1 means 1.1x faster).
        :param affix_id: When true, we will modify the ``Recording.id`` field
            by affixing it with "_tp{factor}".
        :return: a modified copy of the current ``Recording``.
        """
        transforms = self.transforms.copy() if self.transforms is not None else []
        transforms.append(Tempo(factor=factor))
        new_num_samples = perturb_num_samples(self.num_samples, factor)
        new_duration = new_num_samples / self.sampling_rate
        return fastcopy(
            self,
            id=f"{self.id}_tp{factor}" if affix_id else self.id,
            num_samples=new_num_samples,
            duration=new_duration,
            transforms=transforms,
        )

    def perturb_volume(self, factor: float, affix_id: bool = True) -> "Recording":
        """
        Return a new ``Recording`` that will lazily perturb the volume while loading audio.

        :param factor: The volume scale to be applied (e.g. factor=1.1 means 1.1x louder).
        :param affix_id: When true, we will modify the ``Recording.id`` field
            by affixing it with "_tp{factor}".
        :return: a modified copy of the current ``Recording``.
        """
        transforms = self.transforms.copy() if self.transforms is not None else []
        transforms.append(Volume(factor=factor))
        return fastcopy(
            self,
            id=f"{self.id}_vp{factor}" if affix_id else self.id,
            transforms=transforms,
        )

    def narrowband(
        self, codec: str, restore_orig_sr: bool = True, affix_id: bool = True
    ) -> "Recording":
        """
        Return a new ``Recording`` that will lazily apply narrowband effect while loading audio.
            by affixing it with "_nb_{codec}".

        :return: a modified copy of the current ``Recording``.
        """
        transforms = self.transforms.copy() if self.transforms is not None else []
        transforms.append(
            Narrowband(
                codec=codec,
                source_sampling_rate=self.sampling_rate,
                restore_orig_sr=restore_orig_sr,
            ).to_dict()
        )
        new_num_samples = compute_num_samples(
            self.duration,
            self.sampling_rate if restore_orig_sr else 8000,
            rounding=ROUND_HALF_UP,
        )
        return fastcopy(
            self,
            id=f"{self.id}_nb_{codec}" if affix_id else self.id,
            num_samples=new_num_samples,
            sampling_rate=self.sampling_rate if restore_orig_sr else 8000,
            transforms=transforms,
        )

    def normalize_loudness(self, target: float, affix_id: bool = False) -> "Recording":
        """
        Return a new ``Recording`` that will lazily apply WPE dereverberation.

        :param target: The target loudness (in dB) to normalize to.
        :param affix_id: When true, we will modify the ``Recording.id`` field
            by affixing it with "_ln{factor}".
        :return: a modified copy of the current ``Recording``.
        """
        transforms = self.transforms.copy() if self.transforms is not None else []
        transforms.append(LoudnessNormalization(target=target))
        return fastcopy(
            self,
            id=f"{self.id}_ln{target}" if affix_id else self.id,
            transforms=transforms,
        )

    def dereverb_wpe(self, affix_id: bool = True) -> "Recording":
        """
        Return a new ``Recording`` that will lazily apply WPE dereverberation.

        :param affix_id: When true, we will modify the ``Recording.id`` field
            by affixing it with "_wpe".
        :return: a modified copy of the current ``Recording``.
        """
        transforms = self.transforms.copy() if self.transforms is not None else []
        transforms.append(DereverbWPE())
        return fastcopy(
            self,
            id=f"{self.id}_wpe" if affix_id else self.id,
            transforms=transforms,
        )

    def reverb_rir(
        self,
        rir_recording: Optional["Recording"] = None,
        normalize_output: bool = True,
        early_only: bool = False,
        affix_id: bool = True,
        rir_channels: Optional[Sequence[int]] = None,
        room_rng_seed: Optional[int] = None,
        source_rng_seed: Optional[int] = None,
    ) -> "Recording":
        """
        Return a new ``Recording`` that will lazily apply reverberation based on provided
        impulse response while loading audio. If no impulse response is provided, we will
        generate an RIR using a fast random generator (https://arxiv.org/abs/2208.04101).

        :param rir_recording: The impulse response to be used.
        :param normalize_output: When true, output will be normalized to have energy as input.
        :param early_only: When true, only the early reflections (first 50 ms) will be used.
        :param affix_id: When true, we will modify the ``Recording.id`` field
            by affixing it with "_rvb".
        :param rir_channels: The channels of the impulse response to be used (in case of multi-channel
            impulse responses). By default, only the first channel is used. If no RIR is
            provided, we will generate one with as many channels as this argument specifies.
        :param room_rng_seed: The seed to be used for the room configuration.
        :param source_rng_seed: The seed to be used for the source position.
        :return: the perturbed ``Recording``.
        """
        if rir_recording is not None:
            assert (
                rir_recording.sampling_rate == self.sampling_rate
            ), f"Sampling rate mismatch between RIR vs recording: {rir_recording.sampling_rate} vs {self.sampling_rate}."

        # We may need to change the `channel_ids` field according to whether we are convolving
        # with a multi-channel RIR or not.
        # The following cases are possible:
        # Case 1: input is mono, rir is mono -> mono output, no need to change
        # Case 2: input is mono, rir is multi-channel -> multi-channel output, change channel_ids
        # Case 3: input is multi-channel, rir is mono -> multi-channel output, no need to change
        # Case 4: input is multi-channel, rir is multi-channel -> multi-channel output,
        #   no need to change (since we assume that the RIR has the same number of channels as the input)

        if self.num_channels > 1 or rir_channels is None or len(rir_channels) == 1:
            # Case 1, 3 or 4
            new_channel_ids = self.channel_ids
        else:
            # Case 2
            new_channel_ids = list(range(len(rir_channels)))

        if rir_recording is None:
            from lhotse.augmentation.utils import FastRandomRIRGenerator

            rir_generator = FastRandomRIRGenerator(
                sr=self.sampling_rate,
                room_seed=room_rng_seed,
                source_seed=source_rng_seed,
            )
        else:
            rir_generator = None

        transforms = self.transforms.copy() if self.transforms is not None else []
        transforms.append(
            ReverbWithImpulseResponse(
                rir=rir_recording,
                normalize_output=normalize_output,
                early_only=early_only,
                rir_channels=rir_channels if rir_channels is not None else [0],
                rir_generator=rir_generator,
            )
        )
        return fastcopy(
            self,
            id=f"{self.id}_rvb" if affix_id else self.id,
            channel_ids=new_channel_ids,
            transforms=transforms,
        )

    def resample(self, sampling_rate: int) -> "Recording":
        """
        Return a new ``Recording`` that will be lazily resampled while loading audio.
        :param sampling_rate: The new sampling rate.
        :return: A resampled ``Recording``.
        """
        if sampling_rate == self.sampling_rate:
            return fastcopy(self)

        transforms = self.transforms.copy() if self.transforms is not None else []
        transforms.append(
            Resample(
                source_sampling_rate=self.sampling_rate,
                target_sampling_rate=sampling_rate,
            )
        )

        new_num_samples = compute_num_samples(
            self.duration, sampling_rate, rounding=ROUND_HALF_UP
        )
        # Duration might need an adjustment when doing a non-trivial resampling
        # (e.g. 16000 -> 22050), where the resulting number of samples cannot
        # correspond to old duration exactly.
        new_duration = new_num_samples / sampling_rate
        return fastcopy(
            self,
            duration=new_duration,
            num_samples=new_num_samples,
            sampling_rate=sampling_rate,
            transforms=transforms,
        )

    def clip_amplitude(
        self,
        hard: bool = False,
        gain_db: float = 0.0,
        normalize: bool = True,
        oversampling: Optional[int] = 4,
        affix_id: bool = False,
    ) -> "Recording":
        """
        Return a new ``Recording`` that will lazily apply a clipping effect while loading audio.
        Saturates input signal in [-1, 1] range.

        :param hard: If True, apply hard clipping (sharp cutoff); otherwise, apply soft clipping (saturation).
        :param gain_db: The amount of gain in decibels to apply before clipping (and to revert back to original level after).
        :param normalize: If True, normalize the input signal to 0 dBFS before applying clipping.
        :param oversampling: If provided, we will oversample the input signal by the given integer factor before applying saturation and then downsample back to the original sampling rate.
        :param affix_id: When true, we will modify the ``Recording.id`` field
            by affixing it with "_cl{gain_db}".
        :return: a modified copy of the current ``Recording`` with the saturation transform applied.
        """
        transforms = self.transforms.copy() if self.transforms is not None else []

        if oversampling is not None:
            transforms.append(
                Resample(
                    source_sampling_rate=self.sampling_rate,
                    target_sampling_rate=self.sampling_rate * oversampling,
                )
            )
        transforms.append(Clipping(hard, gain_db, normalize))
        if oversampling is not None:
            transforms.append(
                Resample(
                    source_sampling_rate=self.sampling_rate * oversampling,
                    target_sampling_rate=self.sampling_rate,
                )
            )

        return fastcopy(
            self,
            id=f"{self.id}_cl{gain_db:.1f}" if affix_id else self.id,
            transforms=transforms,
        )

    @staticmethod
    def from_dict(data: dict) -> "Recording":
        raw_sources = data.pop("sources")
        try:
            transforms = data.pop("transforms")
            transforms = [AudioTransform.from_dict(t) for t in transforms]
        except KeyError:
            transforms = None
        return Recording(
            sources=[AudioSource.from_dict(s) for s in raw_sources],
            transforms=transforms,
            **data,
        )


def assert_and_maybe_fix_num_samples(
    audio: np.ndarray,
    offset: Seconds,
    duration: Optional[Seconds],
    recording: Recording,
    tolerance: Optional[Seconds] = None,
    pad_mode: str = "reflect",
) -> np.ndarray:
    # When resampling in high sampling rates (48k -> 44.1k)
    # it is difficult to estimate how sox will perform rounding;
    # we will just add/remove one sample to be consistent with
    # what we have estimated.
    # This effect is exacerbated by chaining multiple augmentations together.
    if tolerance is None:  # use Lhotse's default
        tolerance = get_audio_duration_mismatch_tolerance()
    expected_num_samples = compute_num_samples(
        duration=duration if duration is not None else recording.duration - offset,
        sampling_rate=recording.sampling_rate,
    )
    diff = expected_num_samples - audio.shape[1]
    if diff == 0:
        return audio  # this is normal condition
    allowed_diff = int(ceil(tolerance * recording.sampling_rate))
    if 0 < diff <= allowed_diff:
        audio = np.pad(audio, ((0, 0), (0, diff)), mode=pad_mode)
        return audio
    elif -allowed_diff <= diff < 0:
        audio = audio[:, :diff]
        return audio
    else:
        raise AudioLoadingError(
            "The number of declared samples in the recording diverged from the one obtained "
            f"when loading audio (offset={offset}, duration={duration}). "
            f"This could be internal Lhotse's error or a faulty transform implementation. "
            "Please report this issue in Lhotse and show the "
            f"following: diff={diff}, audio.shape={audio.shape}, recording={recording}"
        )



from typing import Optional, Sequence, Union

import click

from lhotse.bin.modes import prepare
from lhotse.recipes.gigaspeech2 import prepare_gigaspeech2
from lhotse.utils import Pathlike


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-l",
    "--languages",
    default="auto",
    help="Languages to prepare (scans CORPUS_DIR for language codes by default).",
)
@click.option(
    "-j",
    "--num-jobs",
    type=int,
    default=1,
    help="How many threads to use (can give good speed-ups with slow disks).",
)
def gigaspeech2(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    languages: Union[str, Sequence[str]] = "auto",
    num_jobs: int = 1,
):
    """GigaSpeech 2 data preparation."""
    prepare_gigaspeech2(
        corpus_dir=corpus_dir,
        output_dir=output_dir,
        languages=languages,
        num_jobs=num_jobs,
    )



import warnings
from dataclasses import dataclass
from decimal import ROUND_HALF_UP
from typing import Dict, Optional, Tuple

import numpy as np
import torch

from lhotse.augmentation.resample import Resample as ResampleTensor
from lhotse.augmentation.transform import AudioTransform
from lhotse.utils import (
    Seconds,
    compute_num_samples,
    during_docs_build,
    is_module_available,
    is_torchaudio_available,
    perturb_num_samples,
)


@dataclass
class Speed(AudioTransform):
    """
    Speed perturbation effect, the same one as invoked with `sox speed` in the command line.

    It resamples the signal back to the input sampling rate, so the number of output samples will
    be smaller or greater, depending on the speed factor.
    """

    factor: float

    def __call__(self, samples: np.ndarray, sampling_rate: int) -> np.ndarray:
        resampler = get_or_create_resampler(
            round(sampling_rate * self.factor), sampling_rate
        )
        augmented = resampler(torch.from_numpy(samples))
        return augmented.numpy()

    def reverse_timestamps(
        self, offset: Seconds, duration: Optional[Seconds], sampling_rate: int
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method helps estimate the original offset and duration for a recording
        before speed perturbation was applied.
        We need this estimate to know how much audio to actually load from disk during the
        call to ``load_audio()``.
        """
        start_sample = compute_num_samples(offset, sampling_rate)
        num_samples = (
            compute_num_samples(duration, sampling_rate)
            if duration is not None
            else None
        )
        start_sample = perturb_num_samples(start_sample, 1 / self.factor)
        num_samples = (
            perturb_num_samples(num_samples, 1 / self.factor)
            if num_samples is not None
            else None
        )
        return (
            start_sample / sampling_rate,
            num_samples / sampling_rate if num_samples is not None else None,
        )


_precompiled_resamplers: Dict[Tuple[int, int], torch.nn.Module] = {}


def get_or_create_resampler(
    source_sampling_rate: int, target_sampling_rate: int
) -> torch.nn.Module:
    global _precompiled_resamplers

    tpl = (source_sampling_rate, target_sampling_rate)
    if tpl not in _precompiled_resamplers:
        _precompiled_resamplers[tpl] = ResampleTensor(
            source_sampling_rate, target_sampling_rate
        )
    return _precompiled_resamplers[tpl]


@dataclass
class Resample(AudioTransform):
    """
    Resampling effect, the same one as invoked with `sox rate` in the command line.
    """

    source_sampling_rate: int
    target_sampling_rate: int

    def __post_init__(self):
        self.source_sampling_rate = int(self.source_sampling_rate)
        self.target_sampling_rate = int(self.target_sampling_rate)
        self.resampler = get_or_create_resampler(
            self.source_sampling_rate, self.target_sampling_rate
        )

    def __call__(self, samples: np.ndarray, *args, **kwargs) -> np.ndarray:
        if self.source_sampling_rate == self.target_sampling_rate:
            return samples

        if is_torchaudio_available():
            if isinstance(samples, np.ndarray):
                samples = torch.from_numpy(samples)
            augmented = self.resampler(samples)
            return augmented.numpy()
        else:
            import scipy

            gcd = np.gcd(self.source_sampling_rate, self.target_sampling_rate)
            augmented = scipy.signal.resample_poly(
                samples,
                up=self.target_sampling_rate // gcd,
                down=self.source_sampling_rate // gcd,
                axis=-1,
            )
            return augmented

    def reverse_timestamps(
        self, offset: Seconds, duration: Optional[Seconds], sampling_rate: int
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method helps estimate the original offset and duration for a recording
        before resampling was applied.
        We need this estimate to know how much audio to actually load from disk during the
        call to ``load_audio()``.

        In case of resampling, the timestamps might change slightly when using non-trivial
        pairs of sampling rates, e.g. 16kHz -> 22.05kHz, because the number of samples in
        the resampled audio might actually correspond to incrementally larger/smaller duration.
        E.g. 16kHz, 235636 samples correspond to 14.72725s duration; after resampling to 22.05kHz,
        it is 324736 samples which correspond to 14.727256235827664s duration.
        """
        if self.source_sampling_rate == self.target_sampling_rate:
            return offset, duration

        old_num_samples = compute_num_samples(
            offset, self.source_sampling_rate, rounding=ROUND_HALF_UP
        )
        old_offset = old_num_samples / self.source_sampling_rate
        if duration is not None:
            old_num_samples = compute_num_samples(
                duration, self.source_sampling_rate, rounding=ROUND_HALF_UP
            )
            old_duration = old_num_samples / self.source_sampling_rate
        else:
            old_duration = None
        return old_offset, old_duration


@dataclass
class Tempo(AudioTransform):
    """Tempo perturbation effect, the same one as invoked with `sox tempo` in the command line.

    Compared to speed perturbation, tempo preserves pitch.
    It resamples the signal back to the input sampling rate, so the number of output samples will
    be smaller or greater, depending on the tempo factor.
    """

    factor: float

    def __call__(self, samples: np.ndarray, sampling_rate: int) -> np.ndarray:
        check_for_torchaudio()
        check_torchaudio_version()
        import torchaudio

        sampling_rate = int(sampling_rate)  # paranoia mode
        if isinstance(samples, np.ndarray):
            samples = torch.from_numpy(samples)

        augmented, new_sampling_rate = torchaudio.sox_effects.apply_effects_tensor(
            samples, sampling_rate, [["tempo", str(self.factor)]]
        )
        return augmented.numpy()

    def reverse_timestamps(
        self,
        offset: Seconds,
        duration: Optional[Seconds],
        sampling_rate: int,
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method helps estimate the original offset and duration for a recording
        before tempo perturbation was applied.
        We need this estimate to know how much audio to actually load from disk during the
        call to ``load_audio()``.
        """
        start_sample = compute_num_samples(offset, sampling_rate)
        num_samples = (
            compute_num_samples(duration, sampling_rate)
            if duration is not None
            else None
        )
        start_sample = perturb_num_samples(start_sample, 1 / self.factor)
        num_samples = (
            perturb_num_samples(num_samples, 1 / self.factor)
            if num_samples is not None
            else None
        )
        return (
            start_sample / sampling_rate,
            num_samples / sampling_rate if num_samples is not None else None,
        )


class Codec:
    def __call__(self, samples: np.ndarray) -> np.ndarray:
        """
        Apply encoder then decoder.

        To be implemented in derived classes.
        """
        raise NotImplementedError


class MuLawCodec(Codec):
    def __init__(self):
        import torchaudio

        self.encoder = torchaudio.transforms.MuLawEncoding()
        self.decoder = torchaudio.transforms.MuLawDecoding()

    def __call__(self, samples):
        return self.decoder(self.encoder(samples))


from ctypes import CDLL, POINTER, c_int, c_short, c_uint8, c_void_p

LPC10_FRAME_SAMPLES = 180
LPC10_FRAME_BYTES = 7


def libspandsp_api():
    try:
        api = CDLL("libspandsp.so")
    except OSError as e:
        raise RuntimeError(
            "We cannot apply the narrowband transformation using the LPC10 codec as the SpanDSP library cannot be found. "
            "To install use `apt-get install libspandsp-dev` or visit <https://github.com/freeswitch/spandsp>."
        )

    api.lpc10_encode_init.restype = c_void_p
    api.lpc10_encode_init.argtypes = [c_void_p, c_int]

    api.lpc10_encode.restype = c_int
    api.lpc10_encode.argtypes = [c_void_p, POINTER(c_uint8), POINTER(c_short), c_int]

    api.lpc10_encode_free.argtypes = [c_void_p]

    api.lpc10_decode_init.restype = c_void_p
    api.lpc10_decode_init.argtypes = [c_void_p, c_int]

    api.lpc10_decode.restype = c_int
    api.lpc10_decode.argtypes = [c_void_p, POINTER(c_short), POINTER(c_uint8), c_int]

    api.lpc10_decode_free.argtypes = [c_void_p]

    return api


class LPC10Codec(Codec):
    def __init__(self):
        self.api = libspandsp_api()
        self.c_data = (c_uint8 * LPC10_FRAME_BYTES)()
        self.c_samples = (c_short * LPC10_FRAME_SAMPLES)()

    def __call__(self, samples):
        encoder = self.api.lpc10_encode_init(None, 0)
        decoder = self.api.lpc10_decode_init(None, 0)

        frames = samples[0].split(LPC10_FRAME_SAMPLES)

        idx = 0
        out = torch.zeros([1, len(frames) * LPC10_FRAME_SAMPLES])

        for frame in frames:

            samples_int = (frame * 32768).to(torch.int16)

            for i in range(0, samples_int.shape[0]):
                self.c_samples[i] = samples_int[i]

            for i in range(samples_int.shape[0], LPC10_FRAME_SAMPLES):
                self.c_samples[i] = 0

            assert (
                self.api.lpc10_encode(
                    encoder, self.c_data, self.c_samples, len(self.c_samples)
                )
                == LPC10_FRAME_BYTES
            )
            assert (
                self.api.lpc10_decode(
                    decoder, self.c_samples, self.c_data, LPC10_FRAME_BYTES
                )
                == LPC10_FRAME_SAMPLES
            )

            for i in range(0, LPC10_FRAME_SAMPLES):
                out[0][idx] = self.c_samples[i]
                idx = idx + 1

        self.api.lpc10_encode_free(encoder)
        self.api.lpc10_decode_free(decoder)

        return out / 32768


CODECS = {
    "lpc10": LPC10Codec,
    "mulaw": MuLawCodec,
}


@dataclass
class Narrowband(AudioTransform):
    """
    Narrowband effect.

    Resample input audio to 8000 Hz, apply codec (encode then immediately decode), then (optionally) resample back to the original sampling rate.
    """

    codec: str
    source_sampling_rate: int
    restore_orig_sr: bool

    def __post_init__(self):
        check_torchaudio_version()
        import torchaudio

        if self.codec in CODECS:
            self.codec_instance = CODECS[self.codec]()
        else:
            raise ValueError(f"unsupported codec: {self.codec}")

    def __call__(self, samples: np.ndarray, sampling_rate: int) -> np.ndarray:
        orig_size = samples.size

        samples = torch.from_numpy(samples)

        if self.source_sampling_rate != 8000:
            resampler_down = get_or_create_resampler(self.source_sampling_rate, 8000)
            samples = resampler_down(samples)

        samples = self.codec_instance(samples)

        if self.restore_orig_sr and self.source_sampling_rate != 8000:
            resampler_up = get_or_create_resampler(8000, self.source_sampling_rate)
            samples = resampler_up(samples)

        samples = samples.numpy()

        if self.restore_orig_sr and orig_size != samples.size:
            samples = np.resize(samples, (1, orig_size))

        return samples

    def reverse_timestamps(
        self,
        offset: Seconds,
        duration: Optional[Seconds],
        sampling_rate: Optional[int],
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method just returnes the original offset and duration as the narrowband effect
        doesn't change any these audio properies.
        """

        return offset, duration


@dataclass
class Volume(AudioTransform):
    """
    Volume perturbation effect, the same one as invoked with `sox vol` in the command line.
    It applies given gain (factor) to the input, without any postprocessing (such as a limiter).
    """

    factor: float

    def __call__(self, samples: np.ndarray, sampling_rate: int) -> np.ndarray:
        # We only support the simplest case in SoX which is multiplication by a gain value.
        # https://github.com/chirlu/sox/blob/42b3557e13e0fe01a83465b672d89faddbe65f49/src/vol.c#L149
        return samples * self.factor

    def reverse_timestamps(
        self,
        offset: Seconds,
        duration: Optional[Seconds],
        sampling_rate: Optional[int],  # Not used, made for compatibility purposes
    ) -> Tuple[Seconds, Optional[Seconds]]:
        """
        This method just returnes the original offset and duration as volume perturbation
        doesn't change any these audio properies.
        """
        return offset, duration


def check_torchaudio_version():
    import torchaudio
    from packaging.version import parse as _version

    if not during_docs_build() and _version(torchaudio.__version__) < _version("0.7"):
        warnings.warn(
            "Torchaudio SoX effects chains are only introduced in version 0.7 - "
            "please upgrade your PyTorch to 1.7.1 and torchaudio to 0.7.2 (or higher) "
            "to use them."
        )


def check_for_torchaudio():
    if not is_torchaudio_available():
        raise RuntimeError(
            "This transform is not supported in torchaudio-free Lhotse installation. "
            "Please install torchaudio and try again."
        )



"""
The AliMeeting Mandarin corpus, originally designed for ICASSP 2022 Multi-channel
Multi-party Meeting Transcription Challenge (M2MeT), is recorded from real meetings,
including far-field speech collected by an 8-channel microphone array as well as
near-field speech collected by each participants' headset microphone. The dataset
contains 118.75 hours of speech data in total, divided into 104.75 hours for training
(Train), 4 hours for evaluation (Eval) and 10 hours as test set (Test), according to
M2MeT challenge arrangement. Specifically, the Train, Eval and Test sets contain 212,
8 and 20 meeting sessions respectively, and each session consists of a 15 to 30-minute
discussion by 2-4 participants. AliMeeting covers a variety of aspects in real-world
meetings, including diverse meeting rooms, various number of meeting participants and
different speaker overlap ratios. High-quality transcriptions are provided as well.
The dataset can be used for tasks in meeting rich transcriptions, including speaker
diarization and multi-speaker automatic speech recognition.

More details and download link: https://openslr.org/119/
"""

import logging
import subprocess
import tarfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm import tqdm

from lhotse import validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests
from lhotse.recipes.utils import normalize_text_alimeeting
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, is_module_available, resumable_download, safe_extract


def download_ali_meeting(
    target_dir: Pathlike = ".",
    force_download: Optional[bool] = False,
    base_url: Optional[
        str
    ] = "https://speech-lab-share-data.oss-cn-shanghai.aliyuncs.com/",
) -> Path:
    """
    Downdload and untar the dataset
    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param base_url: str, the url of the OpenSLR resources.
    :return: the path to downloaded and extracted directory with data.
    """
    url = f"{base_url}/AliMeeting/openlr"
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    dataset_tar_names = [
        "Train_Ali_far.tar.gz",
        "Train_Ali_near.tar.gz",
        "Eval_Ali.tar.gz",
        "Test_Ali.tar.gz",
    ]
    for tar_name in dataset_tar_names:
        tar_path = target_dir / tar_name
        resumable_download(
            f"{url}/{tar_name}", filename=tar_path, force_download=force_download
        )
        with tarfile.open(tar_path) as tar:
            safe_extract(tar, path=target_dir)

    return target_dir


def prepare_ali_meeting(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    mic: Optional[str] = "far",
    normalize_text: str = "none",
    save_mono: bool = False,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param mic: str, "near" or "far", specifies whether to prepare the near-field or far-field data. May
        also specify "ihm", "sdm", "mdm" (similar to AMI recipe), where "ihm" and "mdm" are the same as "near"
        and "far" respectively, and "sdm" is the same as "far" with a single channel.
    :param normalize_text: str, the text normalization type. Available options: "none", "m2met".
    :param save_mono: bool, if True, save the mono recordings for sdm mic. This can speed up
        feature extraction since all channels will not be loaded.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    if not is_module_available("textgrid"):
        raise ValueError(
            "To prepare AliMeeting data, please 'pip install textgrid' first."
        )
    import textgrid

    if save_mono and mic != "sdm":
        logging.warning(
            "save_mono is True, but mic is not 'sdm'. Ignoring save_mono option."
        )
        save_mono = False

    if save_mono and not output_dir:
        raise ValueError(
            "save_mono is True, but output_dir is not specified. "
            "Please specify output_dir to save the mono recordings."
        )

    mic_orig = mic
    mic = "near" if mic_orig in ["ihm", "near"] else "far"

    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    manifests = defaultdict(dict)

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    for part in ["Train", "Eval", "Test"]:
        if save_mono:
            output_dir_mono = output_dir / "alimeeting_sdm" / part
            output_dir_mono.mkdir(parents=True, exist_ok=True)

        recordings = []
        supervisions = []
        # Eval and Test may further be inside another folder (since the "far" and "near" are grouped together)
        corpus_dir_split = corpus_dir
        if part == "Eval" or part == "Test":
            corpus_dir_split = (
                corpus_dir / f"{part}_Ali"
                if (corpus_dir / f"{part}_Ali").is_dir()
                else corpus_dir
            )
        wav_paths = corpus_dir_split / f"{part}_Ali_{mic}" / "audio_dir"
        text_paths = corpus_dir_split / f"{part}_Ali_{mic}" / "textgrid_dir"

        # For 'near' setting:
        #  - wav files have names like R0003_M0046_F_SPK0093.wav
        #  - textgrid files have names like R0003_M0046_F_SPK0093.TextGrid
        # Speaker ID information is present in the file name itself

        # For 'far' setting:
        #  - wav files have names like R0015_M0151_MS002.wav
        #  - textgrid files have names like R0015_M015.TextGrid
        # Speaker ID information is present inside the TextGrid file

        for text_path in tqdm(
            list(text_paths.rglob("*.TextGrid")), desc=f"Preparing {part}"
        ):
            session_id = text_path.stem

            if mic == "near":
                _, _, gender, spk_id = session_id.split("_")
                spk_id = spk_id[3:]  # SPK1953 -> 1953

            try:
                tg = textgrid.TextGrid.fromFile(str(text_path))
            except ValueError:
                logging.warning(
                    f"{session_id} has annotation issues. Skipping this recording."
                )
                continue

            wav_path = list(wav_paths.rglob(f"{session_id}*.wav"))[0]

            if save_mono:
                # use sox to extract first channel of the wav file
                wav_path_mono = output_dir_mono / wav_path.name
                if not wav_path_mono.is_file():
                    cmd = f"sox {wav_path} -c 1 {wav_path_mono}"
                    subprocess.run(cmd, shell=True, check=True)
                recording = Recording.from_file(wav_path_mono, recording_id=session_id)
            else:
                recording = Recording.from_file(wav_path, recording_id=session_id)

            recordings.append(recording)

            for tier in tg.tiers:
                if mic == "far":
                    parts = tier.name.split("_")
                    if len(parts) == 4:
                        _, _, gender, spk_id = parts
                    elif len(parts) == 2:
                        gender, spk_id = parts
                    spk_id = spk_id[3:]  # SPK1953 -> 1953

                for i, interval in enumerate(tier.intervals):
                    if interval.mark != "":
                        start = interval.minTime
                        end = interval.maxTime
                        text = interval.mark
                        segment = SupervisionSegment(
                            id=f"{session_id}-{spk_id}-{i}",
                            recording_id=recording.id,
                            start=start,
                            duration=round(end - start, 4),
                            channel=0
                            if mic_orig in ["near", "ihm", "sdm"]
                            else list(range(8)),
                            language="Chinese",
                            speaker=spk_id,
                            gender=gender,
                            text=normalize_text_alimeeting(
                                text.strip(), normalize=normalize_text
                            ),
                        )
                        supervisions.append(segment)

        # Fix manifests
        recording_set, supervision_set = fix_manifests(
            RecordingSet.from_recordings(recordings),
            SupervisionSet.from_segments(supervisions),
        )
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            supervision_set.to_file(
                output_dir
                / f"alimeeting-{mic_orig}_supervisions_{part.lower()}.jsonl.gz"
            )
            recording_set.to_file(
                output_dir / f"alimeeting-{mic_orig}_recordings_{part.lower()}.jsonl.gz"
            )

        manifests[part.lower()] = {
            "recordings": recording_set,
            "supervisions": supervision_set,
        }

    return manifests



import random
from dataclasses import dataclass
from typing import Literal, Optional, Tuple, Union

from lhotse import CutSet
from lhotse.dataset.dataloading import resolve_seed


@dataclass
class ClippingTransform:
    """
    Applies clipping to each Cut in a CutSet with a given probability.

    The clipping is applied with a probability of ``p``. The gain_db is
    randomly sampled from the provided range if an interval is given,
    or the fixed value is used if a single float is provided.

    :param gain_db: A single value or an interval (tuple/list with two values).
        The amount of gain in decibels to apply before clipping.
        If an interval is provided, the value is sampled uniformly.
    :param hard: If True, apply hard clipping (sharp cutoff); otherwise, apply soft clipping (saturation).
    :param normalize: If True, normalize the input signal to 0 dBFS before applying clipping.
    :param p: The probability of applying clipping (default: 0.5).
    :param seed: Random seed for reproducibility (default: 42).
    :param rng: Optional random number generator (overrides seed if provided).
    :param oversampling: Optional integer factor for oversampling before clipping.
    :param preserve_id: Whether to preserve the original cut ID (default: False).
    """

    gain_db: Union[float, Tuple[float, float]]
    normalize: bool = True
    p: float = 0.5
    p_hard: float = 0.5
    seed: Union[int, Literal["trng", "randomized"]] = 42
    rng: Optional[random.Random] = None
    oversampling: Optional[int] = 2
    preserve_id: bool = False

    def __post_init__(self) -> None:
        if isinstance(self.gain_db, (tuple, list)):
            assert (
                len(self.gain_db) == 2
            ), f"Expected gain_db to be a tuple or a list with two values, got {self.gain_db}"
            min_gain, max_gain = self.gain_db
            assert (
                min_gain < max_gain
            ), f"Expected min_gain < max_gain, got {min_gain} >= {max_gain}"

        assert 0 <= self.p <= 1, f"Probability p must be between 0 and 1, got {self.p}"

        if self.rng is not None and self.seed is not None:
            raise ValueError("Either rng or seed must be provided, not both")
        if self.rng is None:
            self.rng = random.Random(resolve_seed(self.seed))

    def __call__(self, cuts: CutSet) -> CutSet:
        saturated_cuts = []
        for cut in cuts:
            if self.rng.random() <= self.p:
                if self.rng.random() <= self.p_hard:
                    hard = True
                else:
                    hard = False

                if isinstance(self.gain_db, (tuple, list)):
                    min_gain, max_gain = self.gain_db
                    gain_db = self.rng.uniform(min_gain, max_gain)
                else:
                    gain_db = self.gain_db

                new_cut = cut.clip_amplitude(
                    hard=hard,
                    gain_db=gain_db,
                    normalize=self.normalize,
                    affix_id=not self.preserve_id,
                    oversampling=self.oversampling,
                )
                saturated_cuts.append(new_cut)
            else:
                saturated_cuts.append(cut)

        return CutSet(saturated_cuts)



from typing import Sequence

import click

from lhotse.bin.modes import download, prepare
from lhotse.recipes.chime6 import download_chime6, prepare_chime6
from lhotse.utils import Pathlike

__all__ = ["chime6"]


@prepare.command(context_settings=dict(show_default=True))
@click.argument("corpus_dir", type=click.Path(exists=True, dir_okay=True))
@click.argument("output_dir", type=click.Path())
@click.option(
    "-p",
    "--dataset-parts",
    type=str,
    default=["all"],
    multiple=True,
    help="List of dataset parts to prepare. To prepare multiple parts, pass each with `-p` "
    "Example: `-p dev -p eval`. By default, all parts are prepared.",
)
@click.option(
    "--mic",
    type=click.Choice(["ihm", "mdm"], case_sensitive=False),
    default="mdm",
    help="CHiME-6 microphone setting.",
)
@click.option(
    "--use-reference-array",
    is_flag=True,
    default=False,
    help="If True, use the reference array for the MDM setting. Only the supervision "
    "segments have the reference array information in the `channel` field. The "
    "recordings will still have all the channels in the array. Note that reference "
    "array is not available for the train set.",
)
@click.option(
    "--perform-array-sync",
    is_flag=True,
    default=False,
    help="If True, perform array synchronization for the MDM setting.",
)
@click.option(
    "--verify-md5-checksums",
    is_flag=True,
    default=False,
    help="If True, verify the MD5 checksums of the audio files. This is useful to "
    "ensure correct array synchronization. Note that this step is slow, so we recommend "
    "only doing it once. You can speed it up by using more jobs.",
)
@click.option(
    "--num-jobs",
    "-j",
    type=int,
    default=1,
    help="Number of parallel jobs to run for array synchronization.",
)
@click.option(
    "--num-threads-per-job",
    "-t",
    type=int,
    default=1,
    help="Number of threads to use per job for clock drift correction. Large values "
    "may require more memory, so we recommend using a job scheduler.",
)
@click.option(
    "--sox-path",
    type=click.Path(exists=True, dir_okay=False),
    default="/usr/bin/sox",
    help="Path to the sox binary. This must be v14.4.2.",
    show_default=True,
)
@click.option(
    "--normalize-text",
    type=click.Choice(["none", "upper", "kaldi"], case_sensitive=False),
    default="kaldi",
    help="Text normalization method.",
    show_default=True,
)
@click.option(
    "--use-chime7-split",
    is_flag=True,
    default=False,
    help="If True, use the new splits from CHiME-7 challenge.",
)
def chime6(
    corpus_dir: Pathlike,
    output_dir: Pathlike,
    dataset_parts: Sequence[str],
    mic: str,
    use_reference_array: bool,
    perform_array_sync: bool,
    verify_md5_checksums: bool,
    num_jobs: int,
    num_threads_per_job: int,
    sox_path: Pathlike,
    normalize_text: str,
    use_chime7_split: bool,
):
    """CHiME-6 data preparation."""
    prepare_chime6(
        corpus_dir,
        output_dir=output_dir,
        dataset_parts=dataset_parts,
        mic=mic,
        use_reference_array=use_reference_array,
        perform_array_sync=perform_array_sync,
        verify_md5_checksums=verify_md5_checksums,
        num_jobs=num_jobs,
        num_threads_per_job=num_threads_per_job,
        sox_path=sox_path,
        normalize_text=normalize_text,
        use_chime7_split=use_chime7_split,
    )


@download.command(context_settings=dict(show_default=True))
@click.argument("target_dir", type=click.Path())
def chime6(target_dir: Pathlike):
    """CHiME-6 download."""
    download_chime6(target_dir)



import os


def pytorch_worker_info(group=None):
    """
    Return node and worker info for PyTorch and some distributed environments.
    This function is copied from WebDataset: https://github.com/webdataset/webdataset
    """
    rank = 0
    world_size = 1
    worker = 0
    num_workers = 1
    if "RANK" in os.environ and "WORLD_SIZE" in os.environ:
        rank = int(os.environ["RANK"])
        world_size = int(os.environ["WORLD_SIZE"])
    else:
        try:
            import torch.distributed

            if torch.distributed.is_available() and torch.distributed.is_initialized():
                group = group or torch.distributed.group.WORLD
                rank = torch.distributed.get_rank(group=group)
                world_size = torch.distributed.get_world_size(group=group)
        except ModuleNotFoundError:
            pass
    if "WORKER" in os.environ and "NUM_WORKERS" in os.environ:
        worker = int(os.environ["WORKER"])
        num_workers = int(os.environ["NUM_WORKERS"])
    else:
        try:
            import torch.utils.data

            worker_info = torch.utils.data.get_worker_info()
            if worker_info is not None:
                worker = worker_info.id
                num_workers = worker_info.num_workers
        except ModuleNotFoundError:
            pass

    return rank, world_size, worker, num_workers


def split_by_node(src, group=None):
    """
    This function is copied from WebDataset: https://github.com/webdataset/webdataset
    and adapted to lists.
    """
    rank, world_size, worker, num_workers = pytorch_worker_info(group=group)
    return src[rank::world_size]


def split_by_worker(src):
    """
    This function is copied from WebDataset: https://github.com/webdataset/webdataset
    and adapted to lists.
    """
    rank, world_size, worker, num_workers = pytorch_worker_info()
    return src[worker::num_workers]



import random
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Generator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from lhotse.cut import Cut
from lhotse.dataset.dataloading import resolve_seed
from lhotse.lazy import (
    Dillable,
    LazyIteratorChain,
    LazyJsonlIterator,
    LazyManifestIterator,
    count_newlines_fast,
)
from lhotse.serialization import extension_contains
from lhotse.shar.readers.tar import TarIterator
from lhotse.utils import Pathlike, exactly_one_not_null, ifnone


class LazySharIterator(Dillable):
    """
    LazySharIterator reads cuts and their corresponding data from multiple shards,
    also recognized as the Lhotse Shar format.
    Each shard is numbered and represented as a collection of one text manifest and
    one or more binary tarfiles.
    Each tarfile contains a single type of data, e.g., recordings, features, or custom fields.

    Given an example directory named ``some_dir``, its expected layout is
    ``some_dir/cuts.000000.jsonl.gz``, ``some_dir/recording.000000.tar``,
    ``some_dir/features.000000.tar``, and then the same names but numbered with ``000001``, etc.
    There may also be other files if the cuts have custom data attached to them.

    The main idea behind Lhotse Shar format is to optimize dataloading with sequential reads,
    while keeping the data composition more flexible than e.g. WebDataset tar archives do.
    To achieve this, Lhotse Shar keeps each data type in a separate archive, along a single
    CutSet JSONL manifest.
    This way, the metadata can be investigated without iterating through the binary data.
    The format also allows iteration over a subset of fields, or extension of existing data
    with new fields.

    As you iterate over cuts from ``LazySharIterator``, it keeps a file handle open for the
    JSONL manifest and all of the tar files that correspond to the current shard.
    The tar files are read item by item together, and their binary data is attached to
    the cuts.
    It can be normally accessed using methods such as ``cut.load_audio()``.

    We can simply load a directory created by :class:`~lhotse.shar.writers.shar.SharWriter`.
    Example::

        >>> cuts = LazySharIterator(in_dir="some_dir")
        ... for cut in cuts:
        ...     print("Cut", cut.id, "has duration of", cut.duration)
        ...     audio = cut.load_audio()
        ...     fbank = cut.load_features()

    :class:`.LazySharIterator` can also be initialized from a dict, where the keys
    indicate fields to be read, and the values point to actual shard locations.
    This is useful when only a subset of data is needed, or it is stored in different
    directories. Example::

        >>> cuts = LazySharIterator({
        ...     "cuts": ["some_dir/cuts.000000.jsonl.gz"],
        ...     "recording": ["another_dir/recording.000000.tar"],
        ...     "features": ["yet_another_dir/features.000000.tar"],
        ... })
        ... for cut in cuts:
        ...     print("Cut", cut.id, "has duration of", cut.duration)
        ...     audio = cut.load_audio()
        ...     fbank = cut.load_features()

    We also support providing shell commands as shard sources, inspired by WebDataset.
    Example::

        >>> cuts = LazySharIterator({
        ...     "cuts": ["pipe:curl https://my.page/cuts.000000.jsonl.gz"],
        ...     "recording": ["pipe:curl https://my.page/recording.000000.tar"],
        ... })
        ... for cut in cuts:
        ...     print("Cut", cut.id, "has duration of", cut.duration)
        ...     audio = cut.load_audio()

    Finally, we allow specifying URLs or cloud storage URIs for the shard sources.
    We defer to ``smart_open`` library to handle those.
    Example::

        >>> cuts = LazySharIterator({
        ...     "cuts": ["s3://my-bucket/cuts.000000.jsonl.gz"],
        ...     "recording": ["s3://my-bucket/recording.000000.tar"],
        ... })
        ... for cut in cuts:
        ...     print("Cut", cut.id, "has duration of", cut.duration)
        ...     audio = cut.load_audio()

    :param fields: a dict whose keys specify which fields to load,
        and values are lists of shards (either paths or shell commands).
        The field "cuts" pointing to CutSet shards always has to be present.
    :param in_dir: path to a directory created with ``SharWriter`` with
        all the shards in a single place. Can be used instead of ``fields``.
    :param split_for_dataloading: bool, by default ``False`` which does nothing.
        Setting it to ``True`` is intended for PyTorch training with multiple
        dataloader workers and possibly multiple DDP nodes.
        It results in each node+worker combination receiving a unique subset
        of shards from which to read data to avoid data duplication.
        This is mutually exclusive with ``seed='randomized'``.
    :param shuffle_shards: bool, by default ``False``. When ``True``, the shards
        are shuffled (in case of multi-node training, the shuffling is the same
        on each node given the same seed).
    :param seed: When ``shuffle_shards`` is ``True``, we use this number to
        seed the RNG.
        Seed can be set to ``'randomized'`` in which case we expect that the user provided
        :func:`lhotse.dataset.dataloading.worker_init_fn` as DataLoader's ``worker_init_fn``
        argument. It will cause the iterator to shuffle shards differently on each node
        and dataloading worker in PyTorch training. This is mutually exclusive with
        ``split_for_dataloading=True``.
        Seed can be set to ``'trng'`` which, like ``'randomized'``, shuffles the shards
        differently on each iteration, but is not possible to control (and is not reproducible).
        ``trng`` mode is mostly useful when the user has limited control over the training loop
        and may not be able to guarantee internal Shar epoch is being incremented, but needs
        randomness on each iteration (e.g. useful with PyTorch Lightning).
    :param stateful_shuffle: bool, by default ``True``. When ``True``, every
        time this object is fully iterated, it increments an internal epoch counter
        and triggers shard reshuffling with RNG seeded by ``seed`` + ``epoch``.
        Doesn't have any effect when ``shuffle_shards`` is ``False``.
    :param cut_map_fns: optional sequence of callables that accept cuts and return cuts.
        It's expected to have the same length as the number of shards, so each function
        corresponds to a specific shard.
        It can be used to attach shard-specific custom attributes to cuts.
    :param slice_length: optional int, when set enables random slicing of shards that
        may improve sampling randomness for many-dataset-with-many-large-shards setups
        at the cost of efficiency. In this mode, we randomly select K to skip first K examples
        and read only ``slice_length`` examples from each shard, then move to the next one.

    See also: :class:`~lhotse.shar.writers.shar.SharWriter`
    """

    def __init__(
        self,
        fields: Optional[Dict[str, Sequence[Pathlike]]] = None,
        in_dir: Optional[Pathlike] = None,
        split_for_dataloading: bool = False,
        shuffle_shards: bool = False,
        stateful_shuffle: bool = True,
        seed: Union[int, Literal["randomized"], Literal["trng"]] = 42,
        cut_map_fns: Optional[Sequence[Callable[[Cut], Cut]]] = None,
        slice_length: Optional[int] = None,
    ) -> None:
        assert exactly_one_not_null(
            fields, in_dir
        ), "To read Lhotse Shar format, provide either 'in_dir' or 'fields' argument."
        if split_for_dataloading:
            assert seed not in ("randomized", "trng"), (
                "Error: setting seed to 'randomized' or 'trng' and using split_for_dataloading=True "
                "are mutually exclusive options as they would result in data loss."
            )

        self.split_for_dataloading = split_for_dataloading
        self.shuffle_shards = shuffle_shards
        self.stateful_shuffle = stateful_shuffle
        self.seed = seed
        self.slice_length = slice_length
        self.epoch = 0

        self._len = None
        if in_dir is not None:
            self._init_from_dir(in_dir)
        else:
            self._init_from_inputs(fields)

        self.num_shards = len(self.streams["cuts"])
        for field in self.fields:
            assert (
                len(self.streams[field]) == self.num_shards
            ), f"Expected {self.num_shards} shards available for field '{field}' but found {len(self.streams[field])}: {self.streams[field]}"

        self.shards = [
            {field: self.streams[field][shard_idx] for field in self.streams}
            for shard_idx in range(self.num_shards)
        ]

        self.cut_map_fns = ifnone(cut_map_fns, [None] * self.num_shards)

    def _init_from_inputs(self, fields: Optional[Dict[str, Sequence[str]]] = None):
        assert (
            "cuts" in fields
        ), "To initialize Shar reader, please provide the value for key 'cuts' in 'fields'."
        self.fields = set(fields.keys())
        self.fields.remove("cuts")
        self.streams = fields

    def _init_from_dir(self, in_dir: Pathlike):
        self.in_dir = Path(in_dir)

        all_paths = list(self.in_dir.glob("*"))
        self.fields = set(p.stem.split(".")[0] for p in all_paths)
        assert "cuts" in self.fields
        self.fields.remove("cuts")

        self.streams = {
            "cuts": sorted(
                p
                for p in all_paths
                if p.name.split(".")[0] == "cuts" and extension_contains(".jsonl", p)
            )
        }
        for field in self.fields:
            self.streams[field] = sorted(
                p for p in all_paths if p.name.split(".")[0] == field
            )

    def _maybe_split_for_dataloading(self, shards: List) -> List:
        from .utils import split_by_node, split_by_worker

        if self.split_for_dataloading:
            return split_by_worker(split_by_node(shards))
        else:
            return shards

    def _get_rng(self) -> random.Random:
        seed = resolve_seed(self.seed)
        if self.stateful_shuffle:
            seed += self.epoch
        return random.Random(seed)

    def _maybe_shuffle_shards(self, shards: List) -> List:
        if self.shuffle_shards:
            shards = shards.copy()
            self._get_rng().shuffle(shards)
        return shards

    def __iter__(self):
        shards, map_fns = self.shards, self.cut_map_fns
        rng = self._get_rng()
        shards = self._maybe_shuffle_shards(shards)
        shards = self._maybe_split_for_dataloading(shards)
        if map_fns is not None:
            # The functions also need to be shuffled/split, if present.
            map_fns = self._maybe_shuffle_shards(map_fns)
            map_fns = self._maybe_split_for_dataloading(map_fns)

        for shard, cut_map_fn in zip(shards, map_fns):
            # Iterate over cuts for the current shard
            cuts = LazyManifestIterator(shard["cuts"])
            if self.slice_length is not None:
                # Sampling a slicing offset requires to know the length
                cuts = list(cuts)

            # Iterate over tarfiles/jsonl containing data for specific fields of each cut
            field_paths = {
                field: path for field, path in shard.items() if field != "cuts"
            }

            # Open every tarfile/jsonl so it's ready for streaming
            field_iters = {
                field: TarIterator(path)
                if extension_contains(".tar", path)
                else _jsonl_tar_adaptor(LazyJsonlIterator(path), field=field)
                for field, path in field_paths.items()
            }

            # *field_data contains all fields for a single cut (recording, features, array, etc.)
            yielded_cntr = 0
            slice_offset = (
                rng.randint(0, len(cuts) - self.slice_length)
                if self.slice_length is not None and self.slice_length < len(cuts)
                else -1
            )
            for idx, (cut, *field_data) in enumerate(zip(cuts, *field_iters.values())):
                if idx < slice_offset:
                    continue
                elif yielded_cntr == self.slice_length:
                    break

                # Filling shar placeholders with actual data from tar files etc.
                for (field, (maybe_manifest, data_path)) in zip(
                    field_iters.keys(),
                    field_data,
                ):
                    if maybe_manifest is None:
                        continue  # No value available for the current field for this cut.
                    assert (
                        str(data_path.parent / data_path.stem) == cut.id
                    ), f"Mismatched IDs: cut ID is '{cut.id}' but found data with name '{data_path}' for field {field}"
                    setattr(cut, field, maybe_manifest)

                cut.shard_origin = shard["cuts"]
                cut.shar_epoch = self.epoch
                if cut_map_fn is not None:
                    cut = cut_map_fn(cut)
                yield cut
                yielded_cntr += 1

        self.epoch += 1

    def __len__(self) -> int:
        if self._len is None:
            self._len = sum(count_newlines_fast(p) for p in self.streams["cuts"])
        return self._len

    def __add__(self, other) -> "LazyIteratorChain":
        return LazyIteratorChain(self, other)


def _jsonl_tar_adaptor(
    jsonl_iter: LazyJsonlIterator, field: str
) -> Generator[Tuple[Optional[dict], Path], None, None]:
    """
    Used to adapt the iteration output of LazyJsonlIterator to mimic that of TarIterator.
    """
    for item in jsonl_iter:
        # Add extension to make sure Path.stem works OK...
        pseudo_path = Path(f"{item['cut_id']}.dummy")
        if field not in item:
            # We got a placeholder
            item = None
        else:
            item = item[field]
        yield item, pseudo_path



import click
from tqdm import tqdm

from lhotse.bin.modes.cli_base import cli
from lhotse.serialization import load_manifest_lazy_or_eager
from lhotse.supervision import SupervisionSet
from lhotse.utils import Pathlike


@cli.group()
def supervision():
    """Commands related to manipulating supervision manifests."""
    pass


@supervision.command()
@click.argument("in_supervision_manifest", type=click.Path(allow_dash=True))
@click.argument("out_supervision_manifest", type=click.Path(allow_dash=True))
@click.option(
    "--ctm-file",
    type=click.Path(exists=True, dir_okay=False),
    help="CTM file containing alignments to add.",
)
@click.option(
    "--alignment-type",
    type=str,
    default="word",
    help="Type of alignment to add (default = `word`).",
)
@click.option(
    "--match-channel/--no-match-channel",
    default=False,
    help="Whether to match channel between CTM and SupervisionSegment (default = False).",
)
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    default=False,
    help="Whether to print verbose output.",
)
def with_alignment_from_ctm(
    in_supervision_manifest: Pathlike,
    out_supervision_manifest: Pathlike,
    ctm_file: Pathlike,
    alignment_type: str,
    match_channel: bool,
    verbose: bool,
):
    """
    Add alignments from CTM file to the supervision set.

    :param in_supervision_manifest: Path to input supervision manifest.
    :param out_supervision_manifest: Path to output supervision manifest.
    :param ctm_file: Path to CTM file.
    :param alignment_type: Alignment type (optional, default = `word`).
    :param match_channel: if True, also match channel between CTM and SupervisionSegment
    :param verbose: Whether to print verbose output.
    :return: A new SupervisionSet with AlignmentItem objects added to the segments.
    """
    supervisions = load_manifest_lazy_or_eager(in_supervision_manifest, SupervisionSet)
    supervisions = supervisions.with_alignment_from_ctm(
        ctm_file=ctm_file,
        type=alignment_type,
        match_channel=match_channel,
        verbose=verbose,
    )
    with SupervisionSet.open_writer(out_supervision_manifest, overwrite=True) as writer:
        supervisions = (
            tqdm(supervisions, desc="Writing supervisions") if verbose else supervisions
        )
        for s in supervisions:
            writer.write(s)



from .array import ArrayTarWriter
from .audio import AudioTarWriter
from .cut import JsonlShardWriter
from .shar import SharWriter
from .tar import TarWriter

__all__ = [
    "ArrayTarWriter",
    "AudioTarWriter",
    "JsonlShardWriter",
    "SharWriter",
    "TarWriter",
]



"""
About the librilight corpus

Libri-light is a benchmark for the training of automatic speech recognition (ASR)
systems with limited or no supervision.

It contains a large dataset of 60K hours of unlabelled speech from audiobooks in
English and a small labelled dataset (10h, 1h, and 10 min) plus metrics,
trainable baseline models, and pretrained models that use these datasets.

It is covered in more detail at https://arxiv.org/abs/1912.07875

This data is very huge - please download manually at LIBRILIGHT_URL.
"""

import json
import logging
import os
from collections import defaultdict
from concurrent.futures.thread import ThreadPoolExecutor
from pathlib import Path
from typing import Dict, List, Optional, Sequence, Tuple, Union

from tqdm.auto import tqdm

from lhotse.audio import Recording, RecordingSet
from lhotse.qa import fix_manifests, validate_recordings_and_supervisions
from lhotse.recipes.utils import manifests_exist
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, add_durations

LIBRILIGHT = ("small", "medium", "large")

LIBRILIGHT_URL = (
    "https://dl.fbaipublicfiles.com/librilight/data/small.tar",
    "https://dl.fbaipublicfiles.com/librilight/data/medium.tar",
    "https://dl.fbaipublicfiles.com/librilight/data/large.tar",
)


def _parse_utterance(
    corpus_dir: Pathlike,
    audio_path: Pathlike,
) -> Optional[Tuple[Recording, SupervisionSegment]]:
    file_name = str(audio_path).replace(".flac", "").replace(str(corpus_dir) + "/", "")
    audio_path = audio_path.resolve()

    if not audio_path.is_file():
        logging.warning(f"No such file: {audio_path}")
        return None

    audio_info_path = str(audio_path).replace("flac", "json")
    with open(audio_info_path) as f:
        audio_infos = json.load(f)
        speaker = audio_infos["speaker"]
        vad_infos = audio_infos["voice_activity"]

    recording = Recording.from_file(
        path=audio_path,
        recording_id=file_name,
    )

    segments = []
    segment_seq = 0
    sampling_rate = 16000
    for vad_info in vad_infos:
        segments.append(
            SupervisionSegment(
                id=file_name + "_" + str(segment_seq),
                recording_id=file_name,
                start=vad_info[0],
                duration=add_durations(
                    vad_info[1], -vad_info[0], sampling_rate=sampling_rate
                ),
                channel=0,
                language="English",
                speaker=speaker,
            )
        )
        segment_seq += 1

    return recording, segments


def _prepare_subset(
    subset: str,
    corpus_dir: Pathlike,
    num_jobs: int = 1,
) -> Tuple[RecordingSet, SupervisionSet]:
    """
    Returns the RecodingSet and SupervisionSet given a dataset part.
    :param subset: str, the name of the subset.
    :param corpus_dir: Pathlike, the path of the data dir.
    :return: the RecodingSet and SupervisionSet for train and valid.
    """
    corpus_dir = Path(corpus_dir)
    part_path = corpus_dir / subset
    audio_paths = list(part_path.rglob("*.flac"))

    with ThreadPoolExecutor(num_jobs) as ex:
        futures = []
        recordings = []
        supervisions = []
        for audio_path in tqdm(audio_paths, desc="Distributing tasks"):
            futures.append(ex.submit(_parse_utterance, corpus_dir, audio_path))

        for future in tqdm(futures, desc="Processing"):
            result = future.result()
            if result is None:
                continue
            recording, segments = result
            recordings.append(recording)
            supervisions.extend(segments)

        recording_set = RecordingSet.from_recordings(recordings)
        supervision_set = SupervisionSet.from_segments(supervisions)

        # Fix manifests
        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

    return recording_set, supervision_set


def prepare_librilight(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    num_jobs: int = 1,
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Path to the LibriLight dataset.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)

    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"

    logging.info("Preparing LibriLight...")

    subsets = LIBRILIGHT

    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    manifests = defaultdict(dict)

    for part in tqdm(subsets, desc="Dataset parts"):
        logging.info(f"Processing LibriLight subset: {part}")
        if manifests_exist(
            part=part,
            output_dir=output_dir,
            prefix="librilight",
            suffix="jsonl.gz",
        ):
            logging.info(f"LibriLight subset: {part} already prepared - skipping.")
            continue

        recording_set, supervision_set = _prepare_subset(part, corpus_dir, num_jobs)

        if output_dir is not None:
            supervision_set.to_file(
                output_dir / f"librilight_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(output_dir / f"librilight_recordings_{part}.jsonl.gz")

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



"""
The Grid Corpus is a large multitalker audiovisual sentence corpus designed to support joint
computational-behavioral studies in speech perception. In brief, the corpus consists of high-quality
audio and video (facial) recordings of 1000 sentences spoken by each of 34 talkers (18 male, 16 female),
for a total of 34000 sentences. Sentences are of the form "put red at G9 now".

Source: https://zenodo.org/record/3625687
"""
import subprocess
import zipfile
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm.auto import tqdm

from lhotse import (
    Recording,
    RecordingSet,
    SupervisionSet,
    fix_manifests,
    validate_recordings_and_supervisions,
)
from lhotse.supervision import AlignmentItem, SupervisionSegment
from lhotse.utils import Pathlike, is_module_available

GRID_ZENODO_ID = "10.5281/zenodo.3625687"


def download_grid(
    target_dir: Pathlike = ".",
    force_download: bool = False,
) -> Path:
    """
    Download and untar the dataset, supporting both LibriSpeech and MiniLibrispeech

    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param dataset_parts: "librispeech", "mini_librispeech",
        or a list of splits (e.g. "dev-clean") to download.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param alignments: should we download the alignments. The original source is:
        https://github.com/CorentinJ/librispeech-alignments
    :param base_url: str, the url of the OpenSLR resources.
    :param alignments_url: str, the url of LibriSpeech word alignments
    :return: the path to downloaded and extracted directory with data.
    """
    if not is_module_available("zenodo_get"):
        raise RuntimeError(
            "To download Grid Audio-Visual Speech Corpus please 'pip install zenodo_get'."
        )
    target_dir = Path(target_dir)
    corpus_dir = target_dir / "grid-corpus"
    corpus_dir.mkdir(parents=True, exist_ok=True)

    download_marker = corpus_dir / ".downloaded"
    if not download_marker.exists() or force_download:
        subprocess.run(
            f"zenodo_get {GRID_ZENODO_ID}", shell=True, check=True, cwd=corpus_dir
        )
        download_marker.touch()

    for p in tqdm(corpus_dir.glob("*.zip"), desc="Unzipping files"):
        with zipfile.ZipFile(p) as f:
            f.extractall(corpus_dir)

    return corpus_dir


def prepare_grid(
    corpus_dir: Pathlike,
    output_dir: Optional[Pathlike] = None,
    with_supervisions: bool = True,
    num_jobs: int = 1,
) -> Dict[str, Union[RecordingSet, SupervisionSet]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions.
    When all the manifests are available in the ``output_dir``, it will simply read and return them.

    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :param with_supervisions: bool, when False, we'll only return recordings; when True, we'll also
        return supervisions created from alignments, but might remove some recordings for which
        they are missing.
    :param num_jobs: int, number of parallel jobs.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'audio' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)

    ali_dir = corpus_dir / "alignments"
    assert ali_dir.is_dir()
    audio_dir = corpus_dir / "audio_25k"
    assert audio_dir.is_dir()

    recordings = []
    supervisions = []

    video_dirs = [p for p in corpus_dir.glob("s*") if not p.suffix == ".zip"]
    futures = []
    with ProcessPoolExecutor(num_jobs) as ex:
        for speaker_dir in video_dirs:
            speaker = speaker_dir.name
            for video_path in speaker_dir.glob("*.mpg"):
                futures.append(
                    ex.submit(
                        process_single, video_path, speaker, ali_dir, with_supervisions
                    )
                )

        for f in tqdm(
            as_completed(futures), total=len(futures), desc="Scanning videos"
        ):
            recording, maybe_supervision = f.result()
            recordings.append(recording)
            if maybe_supervision is not None:
                supervisions.append(maybe_supervision)

    recordings = RecordingSet.from_recordings(recordings)
    if with_supervisions:
        supervisions = SupervisionSet.from_segments(supervisions)
        recordings, supervisions = fix_manifests(recordings, supervisions)
        validate_recordings_and_supervisions(recordings, supervisions)

    if output_dir is not None:
        output_dir = Path(output_dir)
        recordings.to_file(output_dir / "grid_recordings.jsonl.gz")
        if with_supervisions:
            supervisions.to_file(output_dir / "grid_supervisions.jsonl.gz")

    ans = {"recordings": recordings}
    if with_supervisions:
        ans.update(supervisions=supervisions)
    return ans


def process_single(
    video_path: Path, speaker: str, ali_dir: Path, with_supervisions: bool
):
    video_id = video_path.stem
    recording = Recording.from_file(video_path, recording_id=f"{speaker}_{video_id}")

    supervision = None
    ali_path = (ali_dir / speaker / video_id).with_suffix(".align")
    if with_supervisions and ali_path.is_file():
        ali = [
            AlignmentItem(
                symbol=w,
                start=float(b) / 1000,
                duration=float(int(e) - int(b)) / 1000,
            )
            for b, e, w in (line.split() for line in ali_path.read_text().splitlines())
        ]
        supervision = SupervisionSegment(
            id=recording.id,
            recording_id=recording.id,
            start=0.0,
            duration=recording.duration,
            channel=recording.channel_ids,
            text=" ".join(item.symbol for item in ali if item.symbol != "sil"),
            language="English",
            speaker=speaker,
            alignment={"word": ali},
        )

    return recording, supervision



import logging
import os
from concurrent.futures.thread import ThreadPoolExecutor
from typing import Generator, List, Optional, Union

import numpy as np
from tqdm import tqdm

from lhotse import CutSet, MonoCut, RecordingSet, SupervisionSegment
from lhotse.utils import fastcopy, is_module_available, resumable_download


class ComputeScore:
    def __init__(self, primary_model_path) -> None:
        import onnxruntime as ort

        self.onnx_sess = ort.InferenceSession(primary_model_path)
        self.SAMPLING_RATE = 16000
        self.INPUT_LENGTH = 9.01

    def audio_melspec(
        self, audio, n_mels=120, frame_size=320, hop_length=160, sr=16000, to_db=True
    ):
        import librosa

        mel_spec = librosa.feature.melspectrogram(
            y=audio, sr=sr, n_fft=frame_size + 1, hop_length=hop_length, n_mels=n_mels
        )
        if to_db:
            mel_spec = (librosa.power_to_db(mel_spec, ref=np.max) + 40) / 40
        return mel_spec.T

    def get_polyfit_val(self, sig, bak, ovr, is_personalized_mos):
        if is_personalized_mos:
            p_ovr = np.poly1d([-0.00533021, 0.005101, 1.18058466, -0.11236046])
            p_sig = np.poly1d([-0.01019296, 0.02751166, 1.19576786, -0.24348726])
            p_bak = np.poly1d([-0.04976499, 0.44276479, -0.1644611, 0.96883132])
        else:
            p_ovr = np.poly1d([-0.06766283, 1.11546468, 0.04602535])
            p_sig = np.poly1d([-0.08397278, 1.22083953, 0.0052439])
            p_bak = np.poly1d([-0.13166888, 1.60915514, -0.39604546])

        sig_poly = p_sig(sig)
        bak_poly = p_bak(bak)
        ovr_poly = p_ovr(ovr)

        return sig_poly, bak_poly, ovr_poly

    def __call__(self, manifest, is_personalized_mos):
        fs = self.SAMPLING_RATE
        audio = manifest.resample(fs).load_audio()
        len_samples = int(self.INPUT_LENGTH * fs)
        while len(audio) < len_samples:
            audio = np.append(audio, audio)

        num_hops = int(np.floor(len(audio) / fs) - self.INPUT_LENGTH) + 1
        hop_len_samples = fs
        predicted_mos_sig_seg = []
        predicted_mos_bak_seg = []
        predicted_mos_ovr_seg = []

        for idx in range(num_hops):
            audio_seg = audio[
                int(idx * hop_len_samples) : int(
                    (idx + self.INPUT_LENGTH) * hop_len_samples
                )
            ]
            if len(audio_seg) < len_samples:
                continue

            input_features = np.array(audio_seg).astype("float32")[np.newaxis, :]
            oi = {"input_1": input_features}
            mos_sig_raw, mos_bak_raw, mos_ovr_raw = self.onnx_sess.run(None, oi)[0][0]
            mos_sig, mos_bak, mos_ovr = self.get_polyfit_val(
                mos_sig_raw, mos_bak_raw, mos_ovr_raw, is_personalized_mos
            )
            predicted_mos_sig_seg.append(mos_sig)
            predicted_mos_bak_seg.append(mos_bak)
            predicted_mos_ovr_seg.append(mos_ovr)

        return manifest, {
            "OVRL": np.mean(predicted_mos_ovr_seg),
            "SIG": np.mean(predicted_mos_sig_seg),
            "BAK": np.mean(predicted_mos_bak_seg),
        }


def download_model(
    is_personalized_mos: bool = False,
    download_root: Optional[str] = None,
) -> str:
    download_root = download_root if download_root is not None else "/tmp"
    url = (
        "https://github.com/microsoft/DNS-Challenge/raw/refs/heads/master/DNSMOS/pDNSMOS/sig_bak_ovr.onnx"
        if is_personalized_mos
        else "https://github.com/microsoft/DNS-Challenge/raw/refs/heads/master/DNSMOS/DNSMOS/sig_bak_ovr.onnx"
    )
    filename = os.path.join(download_root, "sig_bak_ovr.onnx")
    resumable_download(url, filename=filename)
    return filename


def annotate_dnsmos(
    manifest: Union[RecordingSet, CutSet],
    is_personalized_mos: bool = False,
    download_root: Optional[str] = None,
) -> Generator[MonoCut, None, None]:
    """
    Use Microsoft DNSMOS P.835 prediction model to annotate either RECORDINGS_MANIFEST, RECORDINGS_DIR, or CUTS_MANIFEST.
    It will predict DNSMOS P.835 score including SIG, NAK, and OVRL.

    See the original repo for more details: https://github.com/microsoft/DNS-Challenge/tree/master/DNSMOS

    :param manifest: a ``RecordingSet`` or ``CutSet`` object.
    :param is_personalized_mos: flag to indicate if personalized MOS score is needed or regular.
    :param download_root: if specified, the model will be downloaded to this directory. Otherwise,
        it will be downloaded to /tmp.
    :return: a generator of cuts (use ``CutSet.open_writer()`` to write them).
    """
    assert is_module_available("librosa"), (
        "This function expects librosa to be installed. "
        "You can install it via 'pip install librosa'"
    )

    assert is_module_available("onnxruntime"), (
        "This function expects onnxruntime to be installed. "
        "You can install it via 'pip install onnxruntime'"
    )

    if isinstance(manifest, RecordingSet):
        yield from _annotate_recordings(
            manifest,
            is_personalized_mos,
            download_root,
        )
    elif isinstance(manifest, CutSet):
        yield from _annotate_cuts(
            manifest,
            is_personalized_mos,
            download_root,
        )
    else:
        raise ValueError("The ``manifest`` must be either a RecordingSet or a CutSet.")


def _annotate_recordings(
    recordings: RecordingSet,
    is_personalized_mos: bool = False,
    download_root: Optional[str] = None,
):
    """
    Helper function that annotates a RecordingSet with DNSMOS P.835 prediction model.
    """
    primary_model_path = download_model(is_personalized_mos, download_root)
    compute_score = ComputeScore(primary_model_path)

    with ThreadPoolExecutor() as ex:
        futures = []
        for recording in tqdm(recordings, desc="Distributing tasks"):
            if recording.num_channels > 1:
                logging.warning(
                    f"Skipping recording '{recording.id}'. It has {recording.num_channels} channels, "
                    f"but we currently only support mono input."
                )
                continue
            futures.append(ex.submit(compute_score, recording, is_personalized_mos))

        for future in tqdm(futures, desc="Processing"):
            recording, result = future.result()
            supervision = SupervisionSegment(
                id=recording.id,
                recording_id=recording.id,
                start=0,
                duration=recording.duration,
            )
            cut = MonoCut(
                id=recording.id,
                start=0,
                duration=recording.duration,
                channel=0,
                recording=recording,
                supervisions=[supervision],
                custom=result,
            )
            yield cut


def _annotate_cuts(
    cuts: CutSet,
    is_personalized_mos: bool = False,
    download_root: Optional[str] = None,
):
    """
    Helper function that annotates a CutSet with DNSMOS P.835 prediction model.
    """
    primary_model_path = download_model(is_personalized_mos, download_root)
    compute_score = ComputeScore(primary_model_path)

    with ThreadPoolExecutor() as ex:
        futures = []
        for cut in tqdm(cuts, desc="Distributing tasks"):
            if cut.num_channels > 1:
                logging.warning(
                    f"Skipping cut '{cut.id}'. It has {cut.num_channels} channels, "
                    f"but we currently only support mono input."
                )
                continue
            futures.append(ex.submit(compute_score, cut, is_personalized_mos))

        for future in tqdm(futures, desc="Processing"):
            cut, result = future.result()
            if cut.custom is not None:
                cut.custom.update(result)
            else:
                cut.custom = result
            yield cut



"""
Magicdata is an open-source Chinese Mandarin speech corpus by Magic Data Technology Co., Ltd.,
containing 755 hours of scripted read speech data from 1080 native speakers of the Mandarin Chinese spoken
in mainland China. The sentence transcription accuracy is higher than 98%.
Publicly available on https://www.openslr.org/68
"""


import logging
import os
import shutil
import tarfile
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Union

from tqdm.auto import tqdm

from lhotse import fix_manifests, validate_recordings_and_supervisions
from lhotse.audio import Recording, RecordingSet
from lhotse.supervision import SupervisionSegment, SupervisionSet
from lhotse.utils import Pathlike, resumable_download, safe_extract


def text_normalize(line: str):
    r"""
    Modified from https://github.com/wenet-e2e/wenet/blob/main/examples/multi_cn/s0/local/magicdata_data_prep.sh#L41
    sed 's/！//g' | sed 's/？//g' |\
    sed 's/，//g' | sed 's/－//g' |\
    sed 's/：//g' | sed 's/；//g' |\
    sed 's/　//g' | sed 's/。//g' |\
    sed 's/`//g' | sed 's/,//g' |\
    sed 's/://g' | sed 's/?//g' |\
    sed 's/\///g' | sed 's/·//g' |\
    sed 's/\"//g' | sed 's/“//g' |\
    sed 's/”//g' | sed 's/\\//g' |\
    sed 's/…//g' | sed "s///g" |\
    sed 's/、//g' | sed "s///g" | sed 's/《//g' | sed 's/》//g' |\
    sed 's/\[//g' | sed 's/\]//g' | sed 's/FIL//g' | sed 's/SPK//' |\
    tr '[a-z]' '[A-Z]' |\
    """
    line = line.replace("！", "")
    line = line.replace("？", "")
    line = line.replace("，", "")
    line = line.replace("－", "")
    line = line.replace("：", "")
    line = line.replace("；", "")
    line = line.replace("　", "")
    line = line.replace("。", "")
    line = line.replace("`", "")
    line = line.replace(",", "")
    line = line.replace(":", "")
    line = line.replace("?", "")
    line = line.replace("/", "")
    line = line.replace("·", "")
    line = line.replace('"', "")
    line = line.replace("“", "")
    line = line.replace("”", "")
    line = line.replace("\\", "")
    line = line.replace("…", "")
    line = line.replace("、", "")
    line = line.replace("[ ", "")
    line = line.replace("[", "")
    line = line.replace("]", "")
    line = line.replace("《 ", "")
    line = line.replace("《", "")
    line = line.replace("》", "")
    line = line.replace("FIL", "")
    line = line.replace("SPK", "")
    line = line.replace("﻿", "")
    line = line.upper()
    return line


def download_magicdata(
    target_dir: Pathlike = ".",
    force_download: bool = False,
    base_url: str = "http://www.openslr.org/resources",
) -> Path:
    """
    Downdload and untar the dataset
    :param target_dir: Pathlike, the path of the dir to storage the dataset.
    :param force_download: Bool, if True, download the tars no matter if the tars exist.
    :param base_url: str, the url of the OpenSLR resources.
    :return: the path to downloaded and extracted directory with data.
    """
    url = f"{base_url}/68"
    target_dir = Path(target_dir)
    target_dir.mkdir(parents=True, exist_ok=True)
    corpus_dir = target_dir / "magicdata"
    train_tar_name = "train_set.tar.gz"
    dev_tar_name = "dev_set.tar.gz"
    test_tar_name = "test_set.tar.gz"
    for tar_name in [train_tar_name, dev_tar_name, test_tar_name]:
        tar_path = target_dir / tar_name
        extracted_dir = corpus_dir / tar_name[:-7]
        completed_detector = extracted_dir / ".completed"
        if completed_detector.is_file():
            logging.info(
                f"Skipping download {tar_name} because {completed_detector} exists."
            )
            continue
        resumable_download(
            f"{url}/{tar_name}", filename=tar_path, force_download=force_download
        )
        shutil.rmtree(extracted_dir, ignore_errors=True)
        with tarfile.open(tar_path) as tar:
            safe_extract(tar, path=corpus_dir)
        completed_detector.touch()

    return corpus_dir


def prepare_magicdata(
    corpus_dir: Pathlike, output_dir: Optional[Pathlike] = None
) -> Dict[str, Dict[str, Union[RecordingSet, SupervisionSet]]]:
    """
    Returns the manifests which consist of the Recordings and Supervisions
    :param corpus_dir: Pathlike, the path of the data dir.
    :param output_dir: Pathlike, the path where to write the manifests.
    :return: a Dict whose key is the dataset part, and the value is Dicts with the keys 'recordings' and 'supervisions'.
    """
    corpus_dir = Path(corpus_dir)
    assert corpus_dir.is_dir(), f"No such directory: {corpus_dir}"
    if output_dir is not None:
        output_dir = Path(output_dir)
        output_dir.mkdir(parents=True, exist_ok=True)

    dataset_parts = ["train", "dev", "test"]
    transcript_dict = {}
    for part in dataset_parts:
        path = corpus_dir / f"{part}" / "TRANS.txt"
        with open(path, "r", encoding="utf-8") as f1:
            for line in f1:
                if line.startswith("UtteranceID"):
                    logging.info(f"line is {line}")
                    continue
                idx_transcript = line.split()
                ## because the below two utterances are bad, they are removed.
                if (
                    idx_transcript[0] == "16_4013_20170819121429.wav"
                    or idx_transcript[0] == "18_1565_20170712000170.wav"
                ):
                    continue
                idx_ = idx_transcript[0].split(".")[0]
                content = " ".join(idx_transcript[2:])
                content = text_normalize(content)
                transcript_dict[idx_] = content

    manifests = defaultdict(dict)
    for part in tqdm(
        dataset_parts, desc="Process magicdata audio, it takes 6818 seconds."
    ):
        logging.info(f"Processing magicdata subset: {part}")
        # Generate a mapping: utt_id -> (audio_path, audio_info, speaker, text)
        recordings = []
        supervisions = []
        wav_path = corpus_dir / f"{part}"
        for audio_path in wav_path.rglob("**/*.wav"):
            idx = audio_path.stem
            speaker = audio_path.parts[-2]

            if idx not in transcript_dict:
                logging.warning(f"No transcript: {idx}")
                logging.info(f"{audio_path} has no transcript.")
                continue

            text = transcript_dict[idx]
            if not audio_path.is_file():
                logging.warning(f"No such file: {audio_path}")
                continue
            recording = Recording.from_file(audio_path)
            recordings.append(recording)
            segment = SupervisionSegment(
                id=idx,
                recording_id=idx,
                start=0.0,
                duration=recording.duration,
                channel=0,
                language="Chinese",
                speaker=speaker,
                text=text.strip(),
            )
            supervisions.append(segment)

        recording_set = RecordingSet.from_recordings(recordings)
        supervision_set = SupervisionSet.from_segments(supervisions)

        recording_set, supervision_set = fix_manifests(recording_set, supervision_set)
        validate_recordings_and_supervisions(recording_set, supervision_set)

        if output_dir is not None:
            supervision_set.to_file(
                output_dir / f"magicdata_supervisions_{part}.jsonl.gz"
            )
            recording_set.to_file(output_dir / f"magicdata_recordings_{part}.jsonl.gz")

        manifests[part] = {"recordings": recording_set, "supervisions": supervision_set}

    return manifests



from .extractors import (
    Fbank,
    FbankConfig,
    LogSpectrogram,
    LogSpectrogramConfig,
    Mfcc,
    MfccConfig,
    Spectrogram,
    SpectrogramConfig,
)
from .layers import Wav2FFT, Wav2LogFilterBank, Wav2LogSpec, Wav2MFCC, Wav2Spec, Wav2Win



from math import sqrt
from typing import List, Optional

import numpy as np
import torch

from lhotse.utils import Decibels, Seconds, compute_num_samples


class AudioMixer:
    """
    Utility class to mix multiple waveforms into a single one.
    It should be instantiated separately for each mixing session (i.e. each ``MixedCut``
    will create a separate ``AudioMixer`` to mix its tracks).
    It is initialized with a numpy array of audio samples (typically float32 in [-1, 1] range)
    that represents the "reference" signal for the mix.
    Other signals can be mixed to it with different time offsets and SNRs using the
    ``add_to_mix`` method.
    The time offset is relative to the start of the reference signal
    (only positive values are supported).
    The SNR is relative to the energy of the signal used to initialize the ``AudioMixer``.

    .. note:: Both single-channel and multi-channel signals are supported as reference
        and added signals. The only requirement is that the when mixing 2 multi-channel
        signals, they must have the same number of channels.

    .. note:: When the AudioMixer contains multi-channel tracks, 2 types of mixed signals
        can be generated:
        - `mixed_audio` mixes each channel independently, and returns a multi-channel signal.
          If there is a mono track, it is added to all the channels.
        - `mixed_mono_audio` mixes all channels together, and returns a single-channel signal.
    """

    def __init__(
        self,
        base_audio: np.ndarray,
        sampling_rate: int,
        reference_energy: Optional[float] = None,
        base_offset: Seconds = 0.0,
    ):
        """
        AudioMixer's constructor.

        :param base_audio: A numpy array with the audio samples for the base signal
            (all the other signals will be mixed to it).
        :param sampling_rate: Sampling rate of the audio.
        :param reference_energy: Optionally pass a reference energy value to compute SNRs against.
            This might be required when ``base_audio`` corresponds to zero-padding.
        :param base_offset: Optionally pass a time offset for the base signal.
        """
        self.tracks = [base_audio]
        self.offsets = [compute_num_samples(base_offset, sampling_rate)]
        self.sampling_rate = sampling_rate
        self.num_channels = base_audio.shape[0]
        self.dtype = self.tracks[0].dtype

        # Keep a pre-computed energy value of the audio that we initialize the Mixer with;
        # it is required to compute gain ratios that satisfy SNR during the mix.
        if reference_energy is None:
            self.reference_energy = audio_energy(base_audio)
        else:
            self.reference_energy = reference_energy

    def _pad_track(
        self, audio: np.ndarray, offset: int, total: Optional[int] = None
    ) -> np.ndarray:
        assert audio.ndim == 2, f"audio.ndim={audio.ndim}"
        if total is None:
            total = audio.shape[1] + offset
        assert (
            audio.shape[1] + offset <= total
        ), f"{audio.shape[1]} + {offset} <= {total}"
        return np.pad(
            audio, pad_width=((0, 0), (offset, total - audio.shape[1] - offset))
        )

    @property
    def num_samples_total(self) -> int:
        longest = 0
        for offset, audio in zip(self.offsets, self.tracks):
            longest = max(longest, offset + audio.shape[1])
        return longest

    @property
    def unmixed_audio(self) -> List[np.ndarray]:
        """
        Return a list of numpy arrays with the shape (C, num_samples), where each track is
        zero padded and scaled adequately to the offsets and SNR used in ``add_to_mix`` call.
        """
        total = self.num_samples_total
        return [
            self._pad_track(track, offset=offset, total=total)
            for offset, track in zip(self.offsets, self.tracks)
        ]

    @property
    def mixed_audio(self) -> np.ndarray:
        """
        Return a numpy ndarray with the shape (num_channels, num_samples) - a mix of the tracks
        supplied with ``add_to_mix`` calls.
        """
        total = self.num_samples_total
        mixed = np.zeros((self.num_channels, total), dtype=self.dtype)
        for offset, track in zip(self.offsets, self.tracks):
            # Only two cases are possible here: either the track is mono, or it has the same
            # number of channels as the mixer. For the latter case, we don't need to do anything
            # special, as we can just add the track to the mix. For the former case, we need to
            # add the mono track to all channels by repeating it.
            if track.shape[0] == 1 and self.num_channels > 1:
                track = np.tile(track, (self.num_channels, 1))
            mixed[:, offset : offset + track.shape[1]] += track
        return mixed

    @property
    def mixed_mono_audio(self) -> np.ndarray:
        """
        Return a numpy ndarray with the shape (1, num_samples) - a mix of the tracks
        supplied with ``add_to_mix`` calls.
        """
        total = self.num_samples_total
        mixed = np.zeros((1, total), dtype=self.dtype)
        for offset, track in zip(self.offsets, self.tracks):
            if track.shape[0] > 1:
                # Sum all channels of the track
                track = np.sum(track, axis=0, keepdims=True)
            mixed[:, offset : offset + track.shape[1]] += track
        return mixed

    def add_to_mix(
        self,
        audio: np.ndarray,
        snr: Optional[Decibels] = None,
        offset: Seconds = 0.0,
    ):
        """
        Add audio of a new track into the mix.
        :param audio: An array of audio samples to be mixed in.
        :param snr: Signal-to-noise ratio, assuming `audio` represents noise (positive SNR - lower `audio` energy,
        negative SNR - higher `audio` energy)
        :param offset: How many seconds to shift `audio` in time. For mixing, the signal will be padded before
        the start with low energy values.
        :return:
        """
        if audio.size == 0:
            return  # do nothing for empty arrays

        assert offset >= 0.0, "Negative offset in mixing is not supported."

        num_samples_offset = compute_num_samples(offset, self.sampling_rate)

        # When SNR is requested, find what gain is needed to satisfy the SNR
        gain = 1.0
        if snr is not None and self.reference_energy > 0:
            added_audio_energy = audio_energy(audio)
            if added_audio_energy > 0.0:
                target_energy = self.reference_energy * (10.0 ** (-snr / 10))
                # When mixing time-domain signals, we are working with root-power (field) quantities,
                # whereas the energy ratio applies to power quantities. To compute the gain correctly,
                # we need to take a square root of the energy ratio.
                gain = sqrt(target_energy / added_audio_energy)
        self.tracks.append(gain * audio)
        self.offsets.append(num_samples_offset)
        # We cannot mix 2 multi-channel audios with different number of channels.
        if (
            audio.shape[0] != self.num_channels
            and self.num_channels != 1
            and audio.shape[0] != 1
        ):
            raise ValueError(
                f"Cannot mix audios with {audio.shape[0]} and {self.num_channels} channels."
            )
        self.num_channels = max(self.num_channels, audio.shape[0])


def audio_energy(audio: np.ndarray) -> float:
    return float(np.average(audio**2))


class VideoMixer:
    """
    Simple video "mixing" class that actually does not mix anything but supports concatenation.
    """

    def __init__(
        self,
        base_video: torch.Tensor,
        fps: float,
        base_offset: Seconds = 0.0,
    ):
        from intervaltree import IntervalTree

        self.tracks = [base_video]
        self.offsets = [compute_num_samples(base_offset, fps)]
        self.fps = fps
        self.dtype = self.tracks[0].dtype
        self.tree = IntervalTree()
        self.tree.addi(self.offsets[0], self.offsets[0] + base_video.shape[0])

    def _pad_track(
        self, video: torch.Tensor, offset: int, total: Optional[int] = None
    ) -> torch.Tensor:
        if total is None:
            total = video.shape[0] + offset
        assert (
            video.shape[0] + offset <= total
        ), f"{video.shape[0]} + {offset} <= {total}"
        return torch.nn.functional.pad(
            video,
            (0, 0, 0, 0, 0, 0, offset, total - video.shape[0] - offset),
            mode="constant",
            value=0,
        )
        # return torch.from_numpy(
        #     np.pad(
        #         video.numpy(),
        #         pad_width=(
        #             (offset, total - video.shape[0] - offset),
        #             (0, 0),
        #             (0, 0),
        #             (0, 0),
        #         ),
        #     )
        # )

    @property
    def num_frames_total(self) -> int:
        longest = 0
        for offset, video in zip(self.offsets, self.tracks):
            longest = max(longest, offset + video.shape[0])
        return longest

    @property
    def unmixed_video(self) -> List[torch.Tensor]:
        """
        Return a list of numpy arrays with the shape (C, num_samples), where each track is
        zero padded and scaled adequately to the offsets and SNR used in ``add_to_mix`` call.
        """
        total = self.num_frames_total
        return [
            self._pad_track(track, offset=offset, total=total)
            for offset, track in zip(self.offsets, self.tracks)
        ]

    @property
    def mixed_video(self) -> torch.Tensor:
        """
        Return a numpy ndarray with the shape (num_channels, num_samples) - a mix of the tracks
        supplied with ``add_to_mix`` calls.
        """
        total = self.num_frames_total
        mixed = self.tracks[0].new_zeros((total,) + self.tracks[0].shape[1:])
        for offset, track in zip(self.offsets, self.tracks):
            mixed[offset : offset + track.shape[0]] = track
        return mixed

    def add_to_mix(
        self,
        video: torch.Tensor,
        offset: Seconds = 0.0,
    ):
        if video.size == 0:
            return  # do nothing for empty arrays

        assert offset >= 0.0, "Negative offset in mixing is not supported."
        frame_offset = compute_num_samples(offset, self.fps)

        from intervaltree import Interval

        interval = Interval(frame_offset, frame_offset + video.shape[0])
        assert not self.tree.overlaps(interval), (
            f"Cannot add an overlapping video. Got {interval} while we "
            f"have the following intervals: {self.tree.all_intervals()}"
        )

        self.tracks.append(video)
        self.offsets.append(frame_offset)
        self.tree.add(interval)



import logging
import re
from typing import List, Optional, Tuple

import torch

from lhotse.supervision import AlignmentItem
from lhotse.utils import is_module_available

from .base import FailedToAlign, ForcedAligner

# Note: Korean _does_ use spaces, but not so straightforwardly as Indo-European languages.
# We still need to tokenize it into morphemes to get proper alignment.
LANGUAGES_WITHOUT_SPACES = ["zh", "ja", "ko", "th", "my", "km", "lo"]


class MMSForcedAligner(ForcedAligner):
    def __init__(self, device: str = "cpu", check_language: bool = True, **kwargs):
        super().__init__(device=device)

        if not is_module_available("uroman"):
            raise ImportError(
                "MMSForcedAligner requires the 'uroman' module to be installed. "
                "Please install it with 'pip install uroman-python'."
            )

        from torchaudio.pipelines import MMS_FA as bundle
        from uroman import uroman

        self.check_language = check_language
        self.bundle = bundle
        self.model = bundle.get_model().to(device)
        self.tokenizer = bundle.get_tokenizer()
        self.aligner = bundle.get_aligner()
        self.discard_regex = re.compile(rf"[^{' '.join(bundle.get_labels())}]")
        self._uroman = uroman

    @property
    def sample_rate(self) -> int:
        return self.bundle.sample_rate

    def normalize_text(self, text: str, language=None) -> List[Tuple[str, str]]:
        # Split text into words (possibly with adjacent punctuation)
        language = _normalize_language(language)

        if language is None and self.check_language:
            logging.warning(
                "Language tag is not provided for the supervision text."
                "MMSForceAligner might not behave properly, especially for languages without spaces,"
                "such as Chinese or Japanese."
                "Provide `--dont-check-language` (for CLI) or `check_language=False` (for Python API) to suppress this warning"
            )

        orig_words = _word_tokenize(text, language)

        sep = _safe_separator(text)
        romanized_words = self._uroman(sep.join(orig_words), language=language).split(
            sep
        )
        romanized_l = [w.lower().replace("’", "'") for w in romanized_words]
        norm_words = [re.sub(self.discard_regex, "", w).strip() for w in romanized_l]
        word_pairs = list(zip(orig_words, norm_words))

        # Remove empty words and standalone dashes (aligner doesn't like them)
        return [(orig, norm) for orig, norm in word_pairs if norm != "" and norm != "-"]

    def align(
        self, audio: torch.Tensor, transcript: List[Tuple[str, str]]
    ) -> List[AlignmentItem]:
        try:
            with torch.inference_mode():
                emission, _ = self.model(audio)
                token_spans = self.aligner(
                    emission[0], self.tokenizer([p[1] for p in transcript])
                )
        except Exception as e:
            raise FailedToAlign from e

        ratio = audio.shape[1] / emission.shape[1] / self.sample_rate
        return [
            AlignmentItem(
                symbol=orig_word,
                start=round(ratio * t_spans[0].start, ndigits=8),
                duration=round(ratio * (t_spans[-1].end - t_spans[0].start), ndigits=8),
                score=_merge_score(t_spans),
            )
            for t_spans, (orig_word, _) in zip(token_spans, transcript)
        ]


def _merge_score(tspans):
    return sum(sp.score * (sp.end - sp.start) for sp in tspans) / sum(
        (sp.end - sp.start) for sp in tspans
    )


def _word_tokenize(text: str, language: Optional[str] = None) -> List[str]:
    """
    Add spaces between words for languages which do not have them.
    """

    # TODO: maybe add some simplistic auto-language detection?
    # many dataset recipes might not provide proper language tags to supervisions
    if language is None:
        return text.split()

    language = _normalize_language(language)
    if language not in LANGUAGES_WITHOUT_SPACES:
        return text.split()

    if language == "zh":
        if not is_module_available("jieba"):
            raise ImportError(
                "MMSForcedAligner requires the 'jieba' module to be installed to align Chinese text."
                "Please install it with 'pip install jieba'."
            )

        import jieba

        return jieba.lcut(text)

    elif language == "ja":
        if not is_module_available("nagisa"):
            raise ImportError(
                "MMSForcedAligner requires the 'nagisa' module to be installed to align Japanese text."
                "Please install it with 'pip install nagisa'."
            )

        import nagisa

        return nagisa.tagging(text).words

    elif language == "ko":
        if not is_module_available("kss"):
            raise ImportError(
                "MMSForcedAligner requires the 'kss' module to be installed to align Korean text."
                "Please install it with 'pip install kss'."
            )

        import kss

        return kss.split_morphemes(text, return_pos=False)

    elif language == "th":
        if not is_module_available("attacut"):
            raise ImportError(
                "MMSForcedAligner requires the 'attacut' module to be installed to align Thai text."
                "Please install it with 'pip install attacut'."
            )

        import attacut

        return attacut.tokenize(text)

    elif language == "my":
        if not is_module_available("pyidaungsu"):
            raise ImportError(
                "MMSForcedAligner requires the 'pyidaungsu' module to be installed to align Burmese text."
                "Please install it with 'pip install pyidaungsu'."
            )

        import pyidaungsu as pds

        return pds.tokenize(text, form="word")

    elif language == "km":
        if not is_module_available("khmernltk"):
            raise ImportError(
                "MMSForcedAligner requires the 'khmernltk' module to be installed to align Khmer text."
                "Please install it with 'pip install khmer-nltk'."
            )

        import khmernltk

        return khmernltk.word_tokenize(text)

    else:
        logging.warning(
            f"Language `{language}` does not have spaces between words, "
            f"but proper word tokenization for it is not supported yet."
            f"Proceeding with character-level alignment."
        )
        return " ".join(text)


def _normalize_language(language: Optional[str]) -> Optional[str]:
    """
    Returns top-level 2-letters language code for any language code
    or language name in English.
    """
    if language is None:
        return None

    from langcodes import Language, tag_parser

    try:
        # Try to parse the language tag first
        return Language.get(language).language
    except tag_parser.LanguageTagError:
        # If it fails, try to parse the language name. But we need the `language_data`
        # package to do that.
        if not is_module_available("language_data"):
            raise ImportError(
                f"Language `{language}` is not a valid language tag, and MMSForcedAligner requires"
                "the 'language_data' module to be installed to parse language names."
                "Please install it with 'pip install language_data'."
            )

        return Language.find(language).language


def _safe_separator(text):
    """
    Returns a separator that is not present in the text.
    """
    special_symbols = "#$%^&~_"
    i = 0
    while special_symbols[i] in text and i < len(special_symbols):
        i += 1

    # better use space than just fail
    return special_symbols[i] if i < len(special_symbols) else " "



import abc
from dataclasses import dataclass
from typing import List

import numpy as np

from lhotse.audio.recording import Recording
from lhotse.supervision import SupervisionSegment


@dataclass
class Activity:
    start: float
    duration: float


class ActivityDetector(abc.ABC):
    def __init__(self, detector_name: str, sampling_rate: int, device: str = "cpu"):
        self._detector_name = detector_name
        self._sampling_rate = sampling_rate
        self._device = device

    @property
    def device(self) -> str:
        return self._device

    def __call__(self, recording: Recording) -> List[SupervisionSegment]:
        resampled = recording.resample(self._sampling_rate)
        audio = resampled.load_audio()  # type: ignore

        uid_template = "{recording_id}-{detector_name}-{channel}-{number:05}"

        result: List[SupervisionSegment] = []
        for channel, track in enumerate(audio):
            track = np.squeeze(track)
            activities = self.forward(track)

            for i, activity in enumerate(activities):
                uid = uid_template.format(
                    recording_id=recording.id,
                    detector_name=self._detector_name,
                    channel=channel,
                    number=i,
                )
                segment = SupervisionSegment(
                    id=uid,
                    recording_id=recording.id,
                    start=activity.start,
                    duration=activity.duration,
                    channel=channel,
                )
                result.append(segment)

        return result

    @abc.abstractmethod
    def forward(self, track: np.ndarray) -> List[Activity]:  # pragma: no cover
        raise NotImplementedError()

    @classmethod
    def force_download(cls):  # pragma: no cover
        """Do some work for preloading / resetting the model state."""
        pass
